/*----------------------------------------------------------------------------*
 * Link.h
 *
 * Michael Geikowski, 2.4.93
 *
 * Definition der Klassen LinkTable und Link zur Verwaltung der zu einem Modul
 * Window gehoerenden Verweise.
 *----------------------------------------------------------------------------*/

#ifndef LINK_H
#define LINK_H

#include "list.h"

class Link : public ListElem
//==========================
	// Speichert Position, Laenge, Klassifizierung und Zielpunkte zu einem
        // Symbol, das Startpunkt von Verweisen ist.
 {
 friend class LinkTable;

 public:

   Link(long pos, int leng, char* text,
        long defpos , char* defmodname , long prevpos, char* prevmodname,
        long nextpos, char* nextmodname, long modpos , char* destmodname )
 //-----------------------------------------------------------------------
 	:SourcePos(pos),      SourceLength(leng),      LinkText(text),
         DefPos(defpos),      DefMod(defmodname),
         PrevUsePos(prevpos), PrevUseMod(prevmodname),
         NextUsePos(nextpos), NextUseMod(nextmodname),
         ModPos(modpos),      Mod(destmodname)
        {}

   ~Link();
 //--------
 	// Loescht LinkText, DefMod, PrevUseMod, NextUseMod, Mod, wenn sie
        // ungleich NULL sind

   void GetSource(long* spos, int* slength) // Position und Laenge des Start-
 //---------------------------------------- // symbols
	{ *spos     = SourcePos;
          *slength  = SourceLength; }

   char* GetLinkText()			    // Klassifizierung des Startsymbols
 //-------------------
        { return LinkText; }

   void GetDef(long* pos, char** ModName)   // Zielpunk : Definition des Symbols
 //--------------------------------------
	{ *pos     = DefPos;
          *ModName = DefMod; }

   void GetPrevUse(long* pos, char** ModName)// Zielpunkt : letzte Verwendung
 //------------------------------------------
	{ *pos     = PrevUsePos;
          *ModName = PrevUseMod; }

   void GetNextUse(long* pos, char** ModName)// Zielpunkt : naechste Verwendung
 //------------------------------------------
	{ *pos     = NextUsePos;
          *ModName = NextUseMod; }

   void GetMod(long* pos, char** ModName)    // Zielpunkt : Modulname
 //--------------------------------------
	{ *pos     = ModPos;
          *ModName = Mod; }

 private:

   char *LinkText;				// Text
   long SourcePos;				// Startsymbol
   int  SourceLength;
   long DefPos,PrevUsePos,NextUsePos,ModPos;	// Zielpunkte
   char *DefMod,*PrevUseMod,*NextUseMod,*Mod;
};

class LinkTable
//=============
	// Verwaltet alle Verweise, die von einem Modul ausgehen
 {
 public:

   LinkTable();
 //------------
 	// Erzeugt leere Tabelle
 
   ~LinkTable();
 //-------------
 
   int ReadLinkFile(char* filename);
 //---------------------------------
	// Liest angegebene Link-Datei ein und erzeugt daraus LinkTable.
	// Ergibt 1, wenn Einlesen fehlerfrei.
        // Ergibt 0, sonst.


   Link* GetLink(long pos);
 //------------------------
	// Liefert Link, der von Position pos ausgeht oder 0, wenn kein
	// solcher Link existiert.
        // Ein Link l geht von pos aus, wenn gilt:
        // Position von l <= pos <= Position von l + Laenge von l

 private:

   List *HashTable;
   int  HashTableLength;
 };

#endif
