/*---------------------------------------------------------------------------*
 * TopWin.c
 *
 * Michael Geikowski, 23.6.93
 *
 * Generierung und Verwaltung des Top Windows
 *---------------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>

#include "adjlistcxx.h"
#include "topwin.h"
#include "semconwin.h"
#include "modwin.h"
#include "navwin.h"
#include "dbinterface.h"

#include <Xm/Xm.h>
#include <Xm/CascadeB.h>
#include <Xm/FileSB.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/List.h>
#include <Xm/PanedW.h>
#include <Xm/PushB.h>
#include <Xm/RowColumn.h>
#include <Xm/ScrollBar.h>
#include <Xm/ScrolledW.h>
#include <Xm/MessageB.h>

// Name des Praeprozessors
//------------------------

#define PREPROC "tpp"

// MenuItem Identifier
//--------------------


#define SMNEW       1
#define SMRESUME    2
#define SMQUIT      3
#define SMABOUT      4
#define GMCNODE    10
#define GMTNODE    11
#define GMMEDGE    12
#define GMNEDGE    13
#define GOBOOKMARK 20
#define GOSEMCON   21

// FSBox Identifier
//-----------------

#define FSBOK     1
#define FSBCANCEL 2

// ScrollBar Identifier
//---------------------

#define SBHOR 1
#define SBVER 2

// Initiale Windowgroesse
//-----------------------

#define CANVASWIDTH         400
#define CANVASHEIGHT        400
#define MSGLISTVISIBLEITEMS   5

// ALIAS - Name fuer einige AdjList Funktionen
//--------------------------------------------

#define GETXPOS()             GetPos()
#define SETXPOS(x)            SetPos(x)
#define GETYPOS()             GetTempPos()
#define SETYPOS(y)            SetTempPos(y)
#define GETWIDGETNR()         GetLevel()
#define SETWIDGETNR(nr)       SetLevel(nr)
#define GETNODEOFFSET()       GetCounter()
#define SETNODEOFFSET(offset) SetCounter(offset)

// Macros
//-------

#define max(a,b) ((a) > (b)) ? (a) : (b) 
#define min(a,b) ((a) < (b)) ? (a) : (b) 

// Modulglobale Variablen
//-----------------------

static int          pipefd[2];		// Kommunikation mit Praeprozessor
static FILE         *fp;
static int          gotmsg;
static int          newSession;
static char         *directory=0;	// dir, name und suffix des aktuellen
static char         *filename=0;	// TopModuls
static char         *suffix=0;
static AdjList      *theGraph=0;	// Graph Darstellung
static GC           GCcopy,GCdash;
static Dimension    FontWidth,FontHeight,HalfFontHeight;
static int          GraphWidth=0;
static int          GraphHeight=0;
static int          CNodeSensitive=1;
static int          TNodeSensitive=1;
static int          MEdgeSolid=1;
static int          NEdgeSolid=1;
static int          x_offset,y_offset;  // Offset fuer ScrollBar Darstellung
static Widget       *WidgetArray=0;	// Buttons fuer Knoten im Graph
static int          WidgetArrayLength;	// Anzahl Knoten
static Widget       FSBox,ScrollBarHor, // global bekannte Widgets
                    ScrollBarVer,canvas,MsgList,NewButton,ResumeButton,
                    GMCButton,GMTButton,GMMButton,GMNButton,
                    BookmarkButton,SemConButton,AboutButton;
static XtAppContext context;		// Merker
static Widget       thetoplevel;

// Schnittstelle zum Praeprozessor
//--------------------------------

static void reinitialise()		// Vorbereitung auf neue Spezifikation
 {					// (alte Spez. vergessen)
 int ok;

 if (filename)				// Lesezeichen retten
   {
   ok = SaveBookmarks(filename);
   if (!ok)
     {
     DisplayMsg("Can't save bookmarks.");
     }
   }
 XmListDeleteAllItems(MsgList);		// Message Area & Graph Area loeschen
 XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,0);
 if (directory) { delete directory; directory = 0; }	// belegten Speicher
 if (filename)  { delete filename;  filename  = 0; }	// freigeben
 if (suffix)    { delete suffix;    suffix    = 0; }
 if (theGraph)  { delete theGraph;  theGraph  = 0; }
 if (WidgetArray)
   {
   for (int i=0; i<WidgetArrayLength; i++)
     {
     XtDestroyWidget(WidgetArray[i]);
     }
   delete [] WidgetArray;
   WidgetArray = 0;
   }
 GraphHeight=0;
 GraphWidth=0;
 DestroyModWins();			// Alte Fenster schliessen
 DestroyNavWin();
 DestroyBookmarkWin();
 DestroySemConWin();
 XtSetSensitive(BookmarkButton,False);
 XtSetSensitive(SemConButton,False);
 }

static void ReadGraph()			// neuen Graphen einlesen aus Datei
 {					// mit Namen filename
 char *name;
 int  x,y,topx,counter,value,ok,endofnodes,endofedges;
 char t;
 Node *n,*m;
 Arg  al[10];
 int  ac;
 XmString  name_str;
 Dimension width,height,maxwidth;

 void NodeCB(Widget w, XtPointer client_data, XtPointer call_data);
 
 if (!OpenGraphFile(filename, "r", 0, 0))	// Oeffne Datei
   {
   DisplayMsg("Can't find graph data.");
   return;
   }
 theGraph = new AdjList;			// neuer Graph
 GetGraphExtension(&GraphWidth,&GraphHeight);	// Graphgroesse lesen
 GraphWidth  = GraphWidth  * FontWidth;		// und merken
 GraphHeight = GraphHeight * FontHeight;
 counter = 0;
 topx    = -1;
 ok = ReadNode(&t, &name, &x, &y, &endofnodes);
 while(!endofnodes && ok)		// Solange nicht alle Knoten gelesen
   {
   n = theGraph->AddNode(name);		// Knoten in AdjList eintragen
   n->SetNodeType(t);
   n->SETXPOS(x * FontWidth);		// setze Knoten Position (Mittelpunkt)
   n->SETYPOS(y * FontHeight);
   if (t != 'D') counter++;		// zaehle alle nicht-dummy Knoten
   if (topx==-1)		        // erster Knoten ist TopModul
     {
     topx = n->GETXPOS();
     }
   ok = ok && ReadEdge(&t, &name, &endofedges);
   while(!endofedges && ok)		// lese alle ausgehenden Kanten
     {
     m = theGraph->AddNode(name);	// Knoten eintragen
     n->AddEdge(m,t);			// Kante eintragen
     ok = ok && ReadEdge(&t, &name, &endofedges);
     }
   ok = ok && ReadNode(&t, &name, &x, &y, &endofnodes);
   }
 CloseGraphFile();
 if (! ok)				// Wenn Lesefehler
   {
   delete theGraph;
   theGraph=0; GraphWidth=0; GraphHeight=0;
   DisplayMsg("Can't read graph data.");
   return;
   }
 WidgetArray = new Widget[counter];	// Array fuer Button-Darstellung der
 WidgetArrayLength = counter;		// Knoten
 counter = 0;
 n = theGraph->GetFirstNode();		// Buttons fuer Knoten erzeugen
 while(n)
   {
   name = n->GetName();
   if (name[0] != '#')			// Wenn kein dummy Knoten
     {
     name_str = XmStringCreateSimple(name);
     ac=0;
     XtSetArg(al[ac],XmNlabelString,name_str); ac++;
     XtSetArg(al[ac],XmNx,10000); ac++;
     XtSetArg(al[ac],XmNy,10000); ac++;
     WidgetArray[counter] = XmCreatePushButton(canvas,"name",al,ac);
     XtManageChild(WidgetArray[counter]);
     XtAddCallback(WidgetArray[counter],XmNactivateCallback,NodeCB,name);
     n->SETWIDGETNR(counter);
     counter++;
     XmStringFree(name_str);
     n->SETNODEOFFSET(HalfFontHeight);	// Abstand von Kantenansatzpunkten
     }					// zu Knotenmittelpunkt
   else	// Wenn dummy Knoten
     {
     n->SETNODEOFFSET(0);
     }
   n = theGraph->GetNextNode();
   }
 ac=0;
 XtSetArg(al[ac],XmNheight,&height); ac++;	// Zeichenflaeche und 
 XtSetArg(al[ac],XmNwidth,&width); ac++;	// Scrollbars init
 XtGetValues(canvas,al,ac);
 y_offset = 0;
 if (width>GraphWidth)
   x_offset = (GraphWidth-width)/2;
 else
   x_offset = topx-width/2;
 maxwidth=max(GraphWidth,width);
 value=min(x_offset,maxwidth-width);
 value=max(0,value);
 ac=0;					// ScrollBars auslegen
 XtSetArg(al[ac],XmNmaximum,maxwidth); ac++;
 XtSetArg(al[ac],XmNsliderSize,width); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,width/2)); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(ScrollBarHor,al,ac);
 ac=0;
 XtSetArg(al[ac],XmNmaximum,max(GraphHeight,height)); ac++;
 XtSetArg(al[ac],XmNsliderSize,height); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,height/2)); ac++;
 XtSetArg(al[ac],XmNvalue,y_offset); ac++;
 XtSetValues(ScrollBarVer,al,ac);
 }

static void ReadPreProc(XtPointer client_data, int *source, XtInputId *id)
	// Zeilenweise die Ausgabe des Praeprozessors lesen
 {
 const int bufinc = 255;
 char *buf,*tempbuf;
 int  i = 0, buflen = bufinc;
 pid_t pid;
 union wait status;

 buf = new char[buflen];
 buf[i] = fgetc(fp);
 while( (buf[i] != '\n') && (! feof(fp)) )
   {
   i++;
   buf[i] = fgetc(fp);
   if (i == buflen-2)			// Puffer bei Ueberlauf vergroessern
     {
     buf[buflen-1] = '\0';
     buflen += bufinc;
     tempbuf = new char[buflen];
     strcpy(tempbuf,buf);
     delete buf;
     buf = tempbuf;
     }
   }
 buf[i] = '\0';
 if (i>0)
   {
   gotmsg = 1;
   DisplayMsg(buf);
   }
 delete buf;
 if (feof(fp))				// Wenn Praeprozessor fertig, kann
   {					// Verbindung aufgeloest und neuer
   fclose(fp);				// Graph gelesen werden
   if (! gotmsg)			// Wenn keine Nachricht von Prae-
     {					// prozessor erhalten
     DisplayMsg("Can't get a connection to the preprocessor.");
     wait(&status);			// Warte bis Praeprzessor terminiert
     }
   else					// mindestens eine Nachricht erhalten
     {
     pid = wait(&status);		// Warte bis Praeprzessor terminiert
     if ( (pid != -1) && (status.w_status == 0) )	// Terminierung ohne
       {						// Fehler
       InitModWins(thetoplevel,directory,suffix);	// ModWin init
       CreateNavWin(thetoplevel);			// NavWin init
       if (!CreateBookmarkWin(thetoplevel,filename))	// bookmarkwin init
         {
         DisplayMsg("Can't read bookmarks.");
	 }
       XtSetSensitive(BookmarkButton,True);

       XtSetSensitive(SemConButton,True);
       
       ReadGraph();					// Graph init
       }
     else				// Terminierung mit Fehler
       {
       DisplayMsg("Preprocessor terminated with error.");
       DisplayMsg("Can't browse this specification.");
       }
     }
   XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
			// expose event wird erzeugt => Graph wird gezeichnet
   XtSetSensitive(NewButton,True);
   XtSetSensitive(ResumeButton,True);
   XtRemoveInput(*id);		// InputEvent loeschen
   }
 }

void CallPreProc()			// Praeprozessor starten und Verbindung
 {					// herstellen
 int pid;
 union wait status;
 char *suffixedfilename;

 if (pipe(pipefd)==-1)
   {
   DisplayMsg("Can't start preprocessor.");
   XtSetSensitive(NewButton,True);
   XtSetSensitive(ResumeButton,True);
   return;
   }
 if ((pid=fork())==0)			// child (=PreProcessor)
   {
   close(pipefd[0]);			// Kind liest nicht aus pipe
   close(2);				// stderr schliessen
   dup(pipefd[1]);			// stderr schreibt jetzt in pipe
   close(pipefd[1]);			// alter Deskriptor nicht mehr noetig
   suffixedfilename = new char[strlen(filename)+strlen(suffix)+2];
   sprintf(suffixedfilename,"%s.%s",filename,suffix);
   execlp(PREPROC,PREPROC,"-n",suffixedfilename,NULL);	// Praeprozessor rufen
   exit(-1);
   }
 else if (pid != -1)			// parent
   {
   close(pipefd[1]);			// Vater schreibt nicht in pipe
   fp = fdopen(pipefd[0],"r");		// pipe Ausgang wird in File fp
					// geleitet
   if (!fp)
     {
     DisplayMsg("Can't get a connection to the preprocessor.");
     DisplayMsg("Please wait...");
     wait(&status);
     DisplayMsg("Preprocessor stopped. Try again.");
     XtSetSensitive(NewButton,True);
     XtSetSensitive(ResumeButton,True);
     return;
     }
   gotmsg = 0;
					// Kommunikation ueber InputEvent
   XtAppAddInput(context,pipefd[0],XtPointer(XtInputReadMask),
                 ReadPreProc,NULL);
   }
 else						// fork() fehlgeschlagen
   {
   DisplayMsg("Can't start preprocessor.");
   XtSetSensitive(NewButton,True);
   XtSetSensitive(ResumeButton,True);
   }
 }



// Callback Funktionen
//--------------------

static void MenuCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Menuitem wurde ausgewaehlt
 {
 int item = (int) client_data;
 XmString info_str1, info_str2, info_str3,info_str4, label_str1, label_str2, label_str ;
 Widget infoButton;
 Arg  al[10];
 int  ac,ok;
 
 XmString  name_str;

 switch (item)
  {
  case SMNEW:   newSession = 1;		// Neue Session beginnen
                XtSetSensitive(NewButton,False);
                XtSetSensitive(ResumeButton,False);
                XtManageChild(FSBox);	// FileSelectionBox oeffnen
                break;
  case SMRESUME:newSession = 0;		// Alte Session fortsetzen
                XtSetSensitive(NewButton,False);
                XtSetSensitive(ResumeButton,False);
                XtManageChild(FSBox);	// FileSelectionBox oeffnen
                break;
  case SMQUIT:  if (filename) SaveBookmarks(filename);
                exit(0);		// Programm beenden
                break;
  case SMABOUT: info_str1 = XmStringCreateLtoR("  tTLA+ Browser, Release 2\n******************************\n",
                                              XmSTRING_DEFAULT_CHARSET);
                info_str2 = XmStringCreateLtoR("Authors:\n   O. Meier (Release 2),\n   M. Geikowski (Release 1),\n",
                                              XmSTRING_DEFAULT_CHARSET);
                label_str1 = XmStringConcat(info_str1,info_str2 ); 
                
                info_str3 = XmStringCreateLtoR("   UNIV. Dortmund, Informatik IV\n\nContact to:\n   A. Mester\n",
                                              XmSTRING_DEFAULT_CHARSET);
                info_str4 = XmStringCreateLtoR("   e-mail: mester@tamfana.informatik.uni-dortmund.de",
                                              XmSTRING_DEFAULT_CHARSET);
                label_str2 = XmStringConcat(info_str3, info_str4);
                label_str = XmStringConcat(label_str1, label_str2 );                   
                ac = 0;
                XtSetArg(al[ac],XmNnoResize, True);ac++;
                XtSetArg(al[ac],XmNdialogType,XmDIALOG_INFORMATION);ac++;
                XtSetArg(al[ac],XmNdialogStyle,XmDIALOG_FULL_APPLICATION_MODAL);ac++;
                XtSetArg(al[ac],XmNtitle,"About tTLA+ Browser...");ac++;
                XtSetArg(al[ac],XmNmessageString,label_str);ac++;
                
                infoButton = XmCreateInformationDialog(AboutButton,"AboutButton",al,ac);
                XtUnmanageChild(XmMessageBoxGetChild(infoButton,XmDIALOG_CANCEL_BUTTON));
                XtUnmanageChild(XmMessageBoxGetChild(infoButton,XmDIALOG_HELP_BUTTON));  
                XtManageChild(infoButton); 
                XmStringFree(info_str1);
                XmStringFree(info_str2);
                XmStringFree(label_str1);
                XmStringFree(info_str3);
                XmStringFree(info_str4);
                XmStringFree(label_str2);
                XmStringFree(label_str);
                break;
  case GOSEMCON : ok = CreateSemConWin(thetoplevel, filename);
                  if(!ok)
                   {
                      DisplayMsg("...can't read variable tree.");
                      XtSetSensitive(SemConButton,False);
                   }                                      
                 break;
  
  
  case GOBOOKMARK:
                DisplayBookmarkWin();   // Lesezeichenfenster anzeigen
                break;
  case GMCNODE: CNodeSensitive = ! CNodeSensitive; // toggle CMOD Darstellung
                if (CNodeSensitive)
                  name_str=XmStringCreateSimple("CONSTANT MODULE sensitive");
                else
                  name_str=XmStringCreateSimple("CONSTANT MODULE insensitive");
                ac=0;
                XtSetArg(al[ac],XmNlabelString,name_str); ac++;
                XtSetValues(GMCButton,al,ac);
                XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// erzeuge expose event
                break;
  case GMTNODE: TNodeSensitive = ! TNodeSensitive; // toggle TMOD Darstellung
                if (TNodeSensitive)
                  name_str=XmStringCreateSimple("TEMPORAL MODULE sensitive");
                else
                  name_str=XmStringCreateSimple("TEMPORAL MODULE insensitive");
                ac=0;
                XtSetArg(al[ac],XmNlabelString,name_str); ac++;
                XtSetValues(GMTButton,al,ac);
                XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// erzeuge expose event
                break;
  case GMMEDGE: MEdgeSolid = ! MEdgeSolid; // toggle IMPORT Darstellung
                if (MEdgeSolid)
                  name_str=XmStringCreateSimple("IMPORT solid");
                else
                  name_str=XmStringCreateSimple("IMPORT dashed");
                ac=0;
                XtSetArg(al[ac],XmNlabelString,name_str); ac++;
                XtSetValues(GMMButton,al,ac);
                XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// erzeuge expose event
                break;
  case GMNEDGE: NEdgeSolid = ! NEdgeSolid; // toggle INCLUDE Darstellung
                if (NEdgeSolid)
                  name_str=XmStringCreateSimple("INCLUDE solid");
                else
                  name_str=XmStringCreateSimple("INCLUDE dashed");
                ac=0;
                XtSetArg(al[ac],XmNlabelString,name_str); ac++;
                XtSetValues(GMNButton,al,ac);
                XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// erzeuge expose event
                break;
  default: assert(0);
  };
 }

static void FileSelectCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Auswahl aus FileSelectionBox auswerten
 {
 XmFileSelectionBoxCallbackStruct* cbp = 
    (XmFileSelectionBoxCallbackStruct*) call_data;
 int ButtonID = (int) client_data;
 char *fullfilename;
 char *p;
 char *slash=0;
 char *dot=0;

 if (XtIsManaged(w))
   {
   XtUnmanageChild(w);
   }
 else
   {
   return;	// Auch bei multiclick darf PreProcessor nur einmal
   }		// gestartet werden.

 if (ButtonID==FSBCANCEL)		// Cancel Button macht nichts
   {
   XtSetSensitive(NewButton,True);
   XtSetSensitive(ResumeButton,True);
   return;
   }
 reinitialise();			// alte Spezifikation loeschen
					// neuen Filename festlegen
 XmStringGetLtoR(cbp->value,XmSTRING_DEFAULT_CHARSET,&fullfilename);
 p = fullfilename;
 while ( *p != '\0' )
   {
   if (*p=='.') { dot=p; }
   if (*p=='/') { dot=0; slash=p; }
   p++;
   }
 if (dot)				// gewaehlten Suffix bestimmen
   {
   *dot = '\0';
   dot++;
   suffix = new char[strlen(dot)+1];
   strcpy(suffix,dot);
   }
 else
   {
   suffix = new char[1];
   suffix[0] = '\0';
   }
 if (slash)				// gewaehltes directory bestimmen
   {
   *slash = '\0';
   slash++;
   directory = new char[strlen(fullfilename)+1];
   strcpy(directory,fullfilename);
   }
 else
   {
   slash = fullfilename;
   directory = new char[2];
   strcpy(directory,".");
   }
 filename = new char[strlen(slash)+1];	// gewaehlten Filenamen bestimmen
 strcpy(filename,slash);
 XtFree(fullfilename);
 chdir(directory);			// wechsle in directory
 if (newSession)
   {
   CallPreProc();			// Praeprozessor mit neuem Filename
					// starten
   }
 else					// alte Session fortsetzen
   {
   InitModWins(thetoplevel,directory,suffix);
   CreateNavWin(thetoplevel);
   if (!CreateBookmarkWin(thetoplevel,filename))
     {
     DisplayMsg("Can't read bookmarks.");
     }
   XtSetSensitive(BookmarkButton,True);
   XtSetSensitive(SemConButton,True);
   ReadGraph();
   XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// expose event wird erzeugt
   XtSetSensitive(NewButton,True);
   XtSetSensitive(ResumeButton,True);
   }
 }

static void ExposeCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Graph neu zeichnen entsprechend aktuellem Bildausschnitt
 {
 XmDrawingAreaCallbackStruct* cbp = (XmDrawingAreaCallbackStruct*) call_data;
 XExposeEvent* event = (XExposeEvent*) cbp->event;
 Node  *n,*m;
 char  t;
 Arg   al[10];
 int   ac;
 int   counter;
 Dimension NodeWidth,NodeHeight;

 if (event->count != 0) return;		// Wenn nicht letzter Callback einer
					// Kette

 if (! theGraph) return;		// Wenn Graph existiert, wird er
					// gezeichnet (entsprechend dem
 counter=0;				// aktuellen Bildausschnitt)
 n = theGraph->GetFirstNode();
 while(n)
   {
   t = n->GetNodeType();		// Knotendarstellung entsprechend
   if (t == 'C')			// Graph Menu
     XtSetSensitive(WidgetArray[counter],CNodeSensitive);
   if (t == 'T')
     XtSetSensitive(WidgetArray[counter],TNodeSensitive);
   if (t != 'D')			// Buttons fuer nicht-dummy Knoten
     {					// positionieren
     ac=0;
     XtSetArg(al[ac],XmNwidth,&NodeWidth); ac++;
     XtSetArg(al[ac],XmNheight,&NodeHeight); ac++;
     XtGetValues(WidgetArray[counter],al,ac);
     XMoveWindow(XtDisplay(WidgetArray[counter]),
                 XtWindow(WidgetArray[counter]),
                 n->GETXPOS()-x_offset-NodeWidth/2,
                 n->GETYPOS()-y_offset-NodeHeight/2);
     counter++;
     }
   m = n->GetFirstEdge();		// Kanten zeichnen in Modus gemaess
   while(m)				// Graph Menu
     {
     t = n->GetTypeOfCurrentEdge();
     if ( ((t == 'M') && MEdgeSolid) || ((t == 'N') && NEdgeSolid))
       XDrawLine(XtDisplay(canvas),XtWindow(canvas),GCcopy,
               n->GETXPOS()-x_offset,n->GETYPOS()-y_offset+n->GETNODEOFFSET(),
               m->GETXPOS()-x_offset,m->GETYPOS()-y_offset-m->GETNODEOFFSET());
     else
       XDrawLine(XtDisplay(canvas),XtWindow(canvas),GCdash,
               n->GETXPOS()-x_offset,n->GETYPOS()-y_offset+n->GETNODEOFFSET(),
               m->GETXPOS()-x_offset,m->GETYPOS()-y_offset-m->GETNODEOFFSET());
     m = n->GetNextEdge();
     }
   n = theGraph->GetNextNode();
   }
 }

static void ResizeCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Zeichenflaeche ist vergroessert worden
 {
 Arg al[10];
 int ac;
 int value;
 Dimension height,width,maxwidth,maxheight;
 Node *n;

 ac=0;					// Scrollbars an neuen Bildausschnitt
 XtSetArg(al[ac],XmNheight,&height); ac++;	// anpassen
 XtSetArg(al[ac],XmNwidth,&width); ac++;
 XtGetValues(canvas,al,ac);
 maxwidth=max(GraphWidth,width);
 if (width>GraphWidth)			// Wenn Zeichenflaeche ganze Graph-
   {					// breite anzeigen kann, soll das
   x_offset = (GraphWidth-width)/2;	// der Knoten des Hauptmodul zentral
   }					// liegen
 value=min(x_offset,maxwidth-width);
 value=max(0,value);
 if (width<=GraphWidth)
   {
   x_offset = value;
   }
 ac=0;					// Scrollbars neu ausrichten
 XtSetArg(al[ac],XmNsliderSize,width); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,width/2)); ac++;
 XtSetArg(al[ac],XmNmaximum,maxwidth); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(ScrollBarHor,al,ac);
 maxheight=max(GraphHeight,height);
 value=min(y_offset,maxheight-height);
 value=max(0,value);
 ac=0;
 XtSetArg(al[ac],XmNsliderSize,height); ac++;
 XtSetArg(al[ac],XmNpageIncrement,max(1,height/2)); ac++;
 XtSetArg(al[ac],XmNmaximum,max(GraphHeight,height)); ac++;
 XtSetArg(al[ac],XmNvalue,value); ac++;
 XtSetValues(ScrollBarVer,al,ac);
 XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// erzeuge expose event
 }

static void ScrollCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Scrollbars wurden bewegt
 {
 Arg al[10];
 int ac;
 int value;
 Dimension width;
 int which = (int) client_data;

 ac=0;					// neuen Bildausschnitt bestimmen
 XtSetArg(al[ac],XmNvalue,&value); ac++;// und zeichnen lassen
 XtGetValues(w,al,ac);
 if (which==SBHOR)			// horizontaler Scrollbar wurde bewegt
   {
   ac=0;
   XtSetArg(al[ac],XmNwidth,&width); ac++;
   XtGetValues(canvas,al,ac);
   if (width<=GraphWidth)
     {
     x_offset=value;
     }
   }
 else					// vertikaler Scrollbar wurde bewegt
   {
   y_offset=value;
   }
 XClearArea(XtDisplay(canvas),XtWindow(canvas),0,0,0,0,1);
					// erzeuge expose event
 }

static void NodeCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Ein Knoten des Graphen wurde angewaehlt
 {
 char *ModName = (char*) client_data;
 Boolean ok;

 ok = OpenModWin(ModName);	// zeige entsprechendes ModWin an
 if (!ok)
   {
   DisplayMsg("Can't open modul window correctly.");
   }
 }

static void ListCB(Widget w, XtPointer client_data, XtPointer call_data)
	// Eine Zeile der MsgArea wurde angewaehlt
 {
 XmListCallbackStruct *cbp = (XmListCallbackStruct*) call_data;

 XmListDeselectPos(MsgList,cbp->item_position);	// highlight abschalten
						// (sonst keine Funktion)
 }

// Hilfsfunktionen zur Window Generierung
//---------------------------------------

static Widget CreateMenu(Widget menubar, char* name, char mnemonic)
	// Ein Menue erzeugen mit Name und Abkuerzung
 {
 Arg al[10];
 int ac;
 Widget   menu,button;
 XmString name_str;

 ac=0;					// menupane erzeugen
 menu = XmCreatePulldownMenu(menubar,name,al,ac);
 name_str = XmStringCreateSimple(name);
 ac=0;					// Button fuer Menu
 XtSetArg(al[ac],XmNsubMenuId,menu); ac++;
 XtSetArg(al[ac],XmNlabelString,name_str); ac++;
 XtSetArg(al[ac],XmNmnemonic,mnemonic); ac++;
 button = XmCreateCascadeButton(menubar,name,al,ac);
 XtManageChild(button);
 XmStringFree(name_str);
 return(menu);
 }

static Widget CreateMenuItem(Widget menupane, char* name, char mnemonic,
                             XtPointer client_data)
	// Menueintrag erzeugen mit Name, Abkuerzung und Daten, die an
	// NodeCB() weitergereicht werden.
 {
 Arg al[10];
 int ac;
 Widget   item;
 XmString name_str;

 name_str = XmStringCreateSimple(name);	// menuitem erzeugen
 ac=0;
 XtSetArg(al[ac],XmNlabelString,name_str); ac++;
 XtSetArg(al[ac],XmNmnemonic,mnemonic); ac++;
 item = XmCreatePushButton(menupane,name,al,ac);
 XtManageChild(item);
 XtAddCallback(item,XmNactivateCallback,MenuCB,client_data);
 XmStringFree(name_str);
 return item;
 }

static void CreateFSBox(Widget parent)
	// FileSelectionBox erzeugen
 {
 Arg al[10];
 int ac;
 char *dir,*tdir;
 Widget button;
 XmString dir_str=NULL;
 XmString pattern_str,title_str;

 dir = getenv("TLASRC");		// default directory fuer TLA+ Modul-
 if (dir)				// Quelltexte
   {
   tdir = new char[strlen(dir)+2];
   sprintf(tdir,"%s/",dir);
   dir_str = XmStringCreateSimple(tdir);
   delete tdir;
   }
 pattern_str = XmStringCreateSimple("*.t");	// default Filter
 title_str = XmStringCreateSimple("tTLA+ Spec Selection");
 ac = 0;
 XtSetArg(al[ac],XmNdialogTitle,title_str); ac++;
 XtSetArg(al[ac],XmNdirectory,dir_str); ac++;
 XtSetArg(al[ac],XmNpattern,pattern_str); ac++;
 FSBox = XmCreateFileSelectionDialog(parent,"FSBox",al,ac);
 XtUnmanageChild(XmFileSelectionBoxGetChild(FSBox,XmDIALOG_HELP_BUTTON));
 button = XmFileSelectionBoxGetChild(FSBox,XmDIALOG_OK_BUTTON);
 ac = 0;
 XtSetArg(al[ac],XmNmultiClick,XmMULTICLICK_DISCARD); ac++;
 XtSetValues(button,al,ac);
 XtAddCallback(FSBox,XmNokCallback,FileSelectCB,XtPointer(FSBOK));
 XtAddCallback(FSBox,XmNcancelCallback,FileSelectCB,XtPointer(FSBCANCEL));
 XmStringFree(pattern_str);
 XmStringFree(title_str);
 if (dir) XmStringFree(dir_str);
 }

static void CreateGC()
	// Graphic Context erzeugen
 {
 Arg al[10];
 int ac;
 Pixel fg,bg;
 XGCValues vals;

 ac=0;					// einer fuer durchgezogene Linien
 XtSetArg(al[ac],XmNforeground,&fg); ac++;
 XtSetArg(al[ac],XmNbackground,&bg); ac++;
 XtGetValues(canvas,al,ac);
 vals.foreground=fg;
 vals.background=bg;
 GCcopy = XtGetGC(canvas,GCForeground | GCBackground,&vals);
					// einer fuer gestrichelte Linien
 vals.foreground=fg;
 vals.background=bg;
 vals.line_style=LineOnOffDash;
 GCdash = XtGetGC(canvas,GCForeground | GCBackground | GCLineStyle,&vals);
 }

static void InitFontConstants()
	// Fonthoehe und -breite merken
 {
 Arg al[10];
 int ac;
 XmString   test_str;
 XmFontList fontlist;

 ac=0;			// Abgucken aus MsgArea
 XtSetArg(al[ac],XmNfontList,&fontlist); ac++;
 XtGetValues(MsgList,al,ac);
 test_str = XmStringCreateSimple("M");
 FontHeight = XmStringHeight(fontlist,test_str);
 FontWidth  = XmStringWidth(fontlist,test_str);
 XmStringFree(test_str);
 HalfFontHeight = FontHeight / 2;
 }

// Exporte
//--------

/*---------------------------------------------------------------------------*
 * Top-Window erzeugen. Top-Window besteht aus Menuzeile, Zeichenflaeche fuer
 * die Darstellung des Modul-Hierarchie-Graphen und einem Message-Window
 *---------------------------------------------------------------------------*/

void CreateTopWin(Widget toplevel, XtAppContext c)
 {
 Arg al[10];
 int ac;
 Widget form,menubar,menu,pane,scrollwin,frame;

 context = c;				// context global merken
 thetoplevel = toplevel;		// toplevel shell merken

 // toplevel->form(1/1)  (i/j) = i-ter von j Soehnen

 ac=0;
 form = XmCreateForm(toplevel,"form",al,ac);
 XtManageChild(form);

 // toplevel->form->menubar(1/2)
 // plus menuitems

 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM); ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM); ac++;
 menubar = XmCreateMenuBar(form,"menubar",al,ac);
 XtManageChild(menubar);
 menu = CreateMenu(menubar,"Session",'S');
   AboutButton = CreateMenuItem(menu,"About",'A',XtPointer(SMABOUT));
   NewButton = CreateMenuItem(menu,"New",'N',XtPointer(SMNEW));
   ResumeButton = CreateMenuItem(menu,"Resume",'R',XtPointer(SMRESUME));
   CreateMenuItem(menu,"Quit",'Q',XtPointer(SMQUIT));
 menu = CreateMenu(menubar,"Goto",'G');
   BookmarkButton = CreateMenuItem(menu,"Bookmarks",'B',XtPointer(GOBOOKMARK));
   XtSetSensitive(BookmarkButton,False);
   SemConButton = CreateMenuItem(menu,"Semantic Concept",'C',XtPointer(GOSEMCON));
   XtSetSensitive(SemConButton, False);
 menu = CreateMenu(menubar,"Graph",'r');
   GMCButton = CreateMenuItem(menu,"CONSTANT MODULE sensitive",'C',
                              XtPointer(GMCNODE));
   GMTButton = CreateMenuItem(menu,"TEMPORAL MODULE sensitive",'T',
                              XtPointer(GMTNODE));
   GMMButton = CreateMenuItem(menu,"IMPORT solid",'M',
                              XtPointer(GMMEDGE));
   GMNButton = CreateMenuItem(menu,"INCLUDE solid",'N',
                              XtPointer(GMNEDGE));

 // toplevel->form->panedwindow(2/2)

 ac=0;
 XtSetArg(al[ac],XmNtopAttachment,XmATTACH_WIDGET);  ac++;
 XtSetArg(al[ac],XmNtopWidget,menubar); 	     ac++;
 XtSetArg(al[ac],XmNleftAttachment,XmATTACH_FORM);   ac++;
 XtSetArg(al[ac],XmNrightAttachment,XmATTACH_FORM);  ac++;
 XtSetArg(al[ac],XmNbottomAttachment,XmATTACH_FORM); ac++;
 pane = XmCreatePanedWindow(form,"pane",al,ac);
 XtManageChild(pane);

 // toplevel->form->panedwindow->scrolledwindow(1/2)

 ac=0;
 XtSetArg(al[ac],XmNscrollBarDisplayPolicy,XmSTATIC); ac++;
 scrollwin = XmCreateScrolledWindow(pane,"scrollwin",al,ac);
 XtManageChild(scrollwin);

 ac=0;
 XtSetArg(al[ac],XmNorientation,XmHORIZONTAL); ac++;
 XtSetArg(al[ac],XmNmaximum,CANVASWIDTH); ac++;
 XtSetArg(al[ac],XmNsliderSize,CANVASWIDTH); ac++;
 XtSetArg(al[ac],XmNpageIncrement,CANVASWIDTH/2); ac++;
 ScrollBarHor = XmCreateScrollBar(scrollwin,"ScrollBarHor",al,ac);
 XtManageChild(ScrollBarHor);
 XtAddCallback(ScrollBarHor,XmNvalueChangedCallback,ScrollCB,XtPointer(SBHOR));

 ac=0;
 XtSetArg(al[ac],XmNorientation,XmVERTICAL); ac++;
 XtSetArg(al[ac],XmNmaximum,CANVASHEIGHT); ac++;
 XtSetArg(al[ac],XmNsliderSize,CANVASHEIGHT); ac++;
 XtSetArg(al[ac],XmNpageIncrement,CANVASHEIGHT/2); ac++;
 ScrollBarVer = XmCreateScrollBar(scrollwin,"ScrollBarVer",al,ac);
 XtManageChild(ScrollBarVer);
 XtAddCallback(ScrollBarVer,XmNvalueChangedCallback,ScrollCB,XtPointer(SBVER));

 ac=0;
 XtSetArg(al[ac],XmNshadowType,XmSHADOW_IN); ac++;
 frame = XmCreateFrame(scrollwin,"frame",al,ac);
 XtManageChild(frame);

 ac=0;
 XtSetArg(al[ac],XmNheight,CANVASHEIGHT); ac++;
 XtSetArg(al[ac],XmNwidth,CANVASWIDTH);  ac++;
 canvas = XmCreateDrawingArea(frame,"canvas",al,ac);
 XtManageChild(canvas);
 XtAddCallback(canvas,XmNexposeCallback,ExposeCB,0);
 XtAddCallback(canvas,XmNresizeCallback,ResizeCB,0);

 XmScrolledWindowSetAreas(scrollwin,ScrollBarHor,ScrollBarVer,frame); 

 // toplevel->form->panedwindow->list(2/2)

 ac=0;
 XtSetArg(al[ac],XmNvisibleItemCount,MSGLISTVISIBLEITEMS); ac++;
 XtSetArg(al[ac],XmNselectionPolicy,XmSINGLE_SELECT); ac++;
 XtSetArg(al[ac],XmNlistSizePolicy,XmRESIZE_IF_POSSIBLE); ac++;
 MsgList = XmCreateScrolledList(pane,"MsgList",al,ac);
 XtManageChild(MsgList);
 XtAddCallback(MsgList,XmNsingleSelectionCallback,ListCB,NULL);
 
 CreateFSBox(toplevel);
 CreateGC();
 InitFontConstants();

 }

 void QuickStart(char* specfile)
 {
  assert(specfile);
  char *fullfilename,*dir;
  char *p;

  char *dot=0;


  fullfilename = new char[strlen(specfile) + 1];
  strcpy( fullfilename, specfile);

  p = fullfilename;
  while ( *p != '\0' )
   {
   if (*p=='.') 
      { dot=p; }
   if (*p=='/') 
      { 
         DisplayMsg("Can't read Specfile with full path !");
         return;
      }      
                
   p++;        
   }

  if (dot)                               // gewaehlten Suffix bestimmen
   {
   *dot = '\0';
   dot++;
   suffix = new char[strlen(dot)+1];
   strcpy(suffix,dot);

   }
  else
   {
   suffix = new char[1];
   suffix[0] = '\0';
   }
 
  dir = getenv("TLASRC");

  if(dir)
   {
     directory = new char[strlen(dir)+1];
     strcpy(directory,dir);
    }
   else
    {
      directory = new char[2];
      strcpy(directory,".");
      }


  filename = new char[strlen(fullfilename)+1];
  strcpy(filename,fullfilename);
  delete fullfilename;

  chdir(directory);
  newSession = 1;
  CallPreProc();
  }

/*---------------------------------------------------------------------------*
 * Meldung im Message-Window anzeigen
 *---------------------------------------------------------------------------*/

void DisplayMsg(char* msg)
 {
 XmString msg_str;

 msg_str = XmStringCreateSimple(msg);
 XmListAddItem(MsgList,msg_str,0);
 XmStringFree(msg_str);
 XmListSetBottomPos(MsgList,0);
 XFlush(XtDisplay(MsgList));
 }
