/*----------------------------------------------------------------------*
 * Env.c
 *
 * Michael Geikowski, 14.6.93
 *
 * Enthaelt die Member-Funktionen der Klasse Environment
 *----------------------------------------------------------------------*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>

#include "env.h"
#include "dbinterface.h"

/*----------------------------------------------------------------------*
 * Hilsklasse Frame
 *----------------------------------------------------------------------*/

class Frame : public ListElem
//===========================
	// Hilfsklasse zur Klasse Environment. Ein Objekt der Klasse Frame
        // beschreibt einen Rahmen einer Umgebung.
 {
 public:

   Frame(Sort_Type defSort, Module_Type t)
 //---------------------------------------
	{ defaultSort = defaultOpSort = defaultParSort = defSort;
	  MType = t;
	  MName = 0;
          RecoveryPoint = 0; }

   ~Frame()
 //--------
	{}

   int Append_help(SymListElem *e, char* actfilename);
 //------------------------------------------------------
	// Hilfsfunktion fuer Environment Append-Funktionen
	// Fuegt den durch `e' beschriebenen Bezeichner in den Frame ein,
        // falls er dort noch nicht existiert.
        // Ergibt 1, wenn e noch nicht existierte oder wenn e existiert aber
        // in einer anderen Datei als actfilename, d.h nicht im Modul
        // bezueglich des globalen Frame oder dessen Untermodulen, definiert
        // wurde. In diesem Fall kann e nur mehrfach in das aktuelle Modul
        // importiert bzw. includiert worden sein, was erlaubt ist.
        // Ergibt 0, sonst, also bei unerlaubtem Konflikt mit bereits
        // definitierten bzw. importierten/includierten Bezeichnern.
  

   SymList     slist;		// Liste der Bezeichner eines Rahmen
   Sort_Type   defaultSort,	// Default-Sorten
               defaultOpSort,
               defaultParSort;
   Module_Type MType;		// Modultyp
   SymListElem *MName;		// Modulname
   int         MDepth;		// Schachtelungstiefe von Untermodulen
   char        RecoveryPoint;	// Fehlerbehebungspunkt
};

int Frame::Append_help(SymListElem* e, char *actfilename)
 {
 SymListElem *d;
 char *dDefMod, *eDefMod;

 d = slist.Lookup(e,e->GetQualifier());	 // Suche e in Frame
 eDefMod = e->GetDefiningModule();
 if (d)			// wenn Bezeichner mit gleichem Namen und gleicher
   {			// Qualifiziererliste schon im Frame existiert
   dDefMod = d->GetDefiningModule();
   assert(dDefMod && eDefMod);
   if ( (strcmp(eDefMod,dDefMod)!=0) || (strcmp(eDefMod,actfilename)==0) )
     {
     return 0;
     }
   }
 else		// e existiert noch nicht
   {
   slist.Append(e);	// e einfuegen
   }
 return 1;
 }

/*----------------------------------------------------------------------*
 * Klasse Environment
 *----------------------------------------------------------------------*/

Environment::Environment(Sort_Type defSort, Module_Type t, char* filename)
 {
 actfilename = filename;
 Frame *p = new Frame(defSort,t);	 // init globalen Rahmen
 NewFirstElem(p);
 p->RecoveryPoint = 1;
 p->MDepth = 1;
 }

Environment::~Environment()
 {
 }

void Environment::Pop()
 {
 DelFirstElem();
 }

void Environment::Push()
 {
 Frame *q = (Frame*) GetFirstElem();

 Frame *p = new Frame(q->defaultParSort,q->MType);
 p->defaultOpSort = q->defaultOpSort;
 p->MName	  = q->MName;
 p->MDepth        = q->MDepth;
 NewFirstElem(p);
 }

void Environment::Append(SymListElem* e)
 {
 assert(e);
 Frame *p = (Frame*) GetFirstElem();	// Top Rahmen
 assert(p);
 int ok = p->Append_help(e,actfilename);// einfuegen
 if (!ok)
   {
   SimpleError("symbol defined twice",e->pos,e->leng);
   }
 }

void Environment::AppendInEnclosingFrame(SymListElem* e)
 {
 assert(e);
 Frame *p = (Frame*) GetFirstElem();	// Top Rahmen
 assert(p);
 p = (Frame*) GetNextElem();		// Top Rahmen umfassender Rahmen
 assert(p);
 int ok = p->Append_help(e,actfilename);// einfuegen
 if (!ok)
   {
   SimpleError("symbol defined twice",e->pos,e->leng);
   }
 }

void Environment::Append(SymList *s, int isIncludeList,
                         SymListElem* Import)
 {
 SymListElem *e;
 char           *errmsg;
 int            ok;
 
 assert(s);
 Frame *p = (Frame*) GetFirstElem();	// Top Rahmen
 assert(p);
 e = s->GetFirstElem();			// Je Element
 while (e)
   {
   if (e->Get_nidt() != nidt_substPar)	// Substituierte Parameter werden nicht
     {					// includiert
     if ( isIncludeList && (e->Get_nidt()==nidt_assumption) )
       {
       e->Set_nidt(nidt_theorem);	// Klasse wandeln
       }
     ok = p->Append_help(e,actfilename);// einfuegen
     if (!ok)
       {
       if (Import)
         {
         errmsg = new char[strlen("imported/included symbol '' defined twice")+
                           strlen(e->Getlexem())+1];
         sprintf(errmsg,"imported/included symbol '%s' defined twice",
                     e->Getlexem());
         SimpleError(errmsg,Import->pos,Import->leng);
         delete errmsg;
         }
       else
         {
         SimpleError("symbol defined twice",e->pos,e->leng);
         }
       }
     }
   e = s->GetNextElem();
   }
 }

void Environment::StartSubModule(Sort_Type defSort, Module_Type t)
 {
 Push();
 Frame *p = (Frame*) GetFirstElem();
 assert(p);
 p->defaultSort = p->defaultOpSort = p->defaultParSort = defSort;
 p->MType = t;
 p->RecoveryPoint = 1;
 (p->MDepth)++;
 }

void Environment::AppendAndPopSubModule()
 {
 int ok;

 Frame *p = (Frame*) GetFirstElem();	// globaler Rahmen des Untermoduls
 assert(p);
 Frame *q = (Frame*) GetNextElem();	// umfassender Rahmen
 assert(q);
 SymListElem *d = p->MName;		// Append
 if (d)					// Erst ModName als qualifier
   {
   d->Set_nidt(nidt_qualifier);
   d->SetSort(new Sort(s_const));
   ok = q->Append_help(d,actfilename);
   if (!ok)
     {
     SimpleError("symbol defined twice",d->pos,d->leng);
     }
   }
 SymListElem *e = p->slist.GetFirstElem();		// Dann restliche
 while (e)						// Symbole evtl.
   {							// qualifiziert
   if (d
         && (e->Get_nidt()!=nidt_formalglobPar)
         && (e->Get_nidt()!=nidt_qualifier)  )
     {
     e->AddQualifier(d);
     }
   ok = q->Append_help(e,actfilename);
   if (!ok)
     {
     SimpleError("symbol defined twice",e->pos,e->leng);
     }
   e = p->slist.GetNextElem();
   }
 DelFirstElem();					// Pop
 }

SymListElem* Environment::Lookup(SymListElem *e, SymList *qual)
 {
 Frame *p;
 SymListElem *q;

 assert(e && qual);
 p = (Frame*) GetFirstElem();		// Fuer alle Rahmen
 while (p)
   {					// Fuer alle Elemente je Rahmen
   q = p->slist.Lookup(e,qual);
   if (q)
     {
     return(q);				// Wenn gefunden, dann melden
     }
   p = (Frame*) GetNextElem();		// Umfassender Rahmen
   }
 return 0;
 }

static void UseOfID_help(SymListElem* f, SymListElem* e, SymList* s)
	// Definition des Bezeichners e im Umgebungsmodell ist Bezeichner f.
	// s enthaelt alle Qualifier von e, e selber und alle Record-
	// Komponenten-Namen, die e angeheftet wurden.
	// e wird auf korrekte Verwendung von f ueberprueft und mit f
	// und weiteren Verwendungen von f verknuepft.
 {
 Sort *sort;
 int  ok,level;

 assert(f->GetSort());
 level = 0;		// Wurde e mit korrekter Sorte verwendet ?
 sort = f->GetSort()->copy();
 if ( (e->GetSort()) && (f->GetSort()->GetLength()==1))
   {
   SimpleError("illegal use of parameterlist",e->pos,e->leng);
   }
 else if ( (e->GetSort()) && (f->GetSort()->GetLength()>1))
   {
   ok = ( f->GetSort()->legal_call(*(e->GetSort())) );
   while ((!ok) && level<3)	// sort-raising
    {
    if (sort) delete sort;
    level++;
    sort=f->GetSort()->raise(level);
    ok = ((sort!=NULL) && sort->legal_call(*(e->GetSort())));
    }
   if (!ok)	// falsche Sorte
     {
     if (sort) delete sort;
     sort = f->GetSort()->copy();
     SimpleError("illegal parameterlist",e->pos,e->leng);
     SimpleErrorCont("  used     : ");
     e->GetSort()->ErrorOutPut(1);
     SimpleErrorCont("\n  expected : ");
     f->GetSort()->ErrorOutPut2();
     SimpleErrorCont("\n");
     }
   }
 if (e->GetSort()) delete e->GetSort();
 e->SetSort(sort);
 e->SetDefinedAt(f);		// Verknuepfung: e mit f
 e->Set_nidt(f->Get_nidt());
 e = s->GetNextElem();
 while(e)		// eventuelle Record-Komponenten-Namen pr\"ufen
   {
   if ( e->GetSort() )
     {
     SimpleError("no parameterlist allowed for record component name",
     e->pos,e->leng);
     }
   e->Set_nidt(nidt_rec_comp);
   e = s->GetNextElem();
   }
 }

SymListElem* Environment::UseOfID(SymList* s)
 {
 SymList *qual  = new SymList;
 SymList *empty = new SymList;
 SymListElem *e = s->GetFirstElem();
 SymListElem *f;

 while(e)	// Durchlaufe s, bis ein nicht-Qualifier gefunden.
   {
   f = Lookup(e,qual);
   if (f && (f->Get_nidt()!=nidt_qualifier))
     {
     break;	// nicht-Qualifier gefunden
     }
   f = Lookup(e,empty); 	// Qualifier haben selbst keine Qualifier
   if (f && (f->Get_nidt()==nidt_qualifier))
     {				// Wenn Qualifier gefunden...
     assert(f->GetSort());	// ... Parameter testen
     if ( (e->GetSort()==0) && (f->GetSort()->GetLength()!=1) )
       {
       SimpleError("missing parameterlist for qualifier",e->pos,e->leng);
       }
     if ( (e->GetSort()) && (! f->GetSort()->legal_call(*(e->GetSort()))))
       {
       SimpleError("illegal parameterlist for qualifier",e->pos,e->leng);
       SimpleErrorCont  ("  used     : ");
       e->GetSort()->ErrorOutPut(1);
       SimpleErrorCont("\n  expected : ");
       f->GetSort()->ErrorOutPut2();
       SimpleErrorCont("\n");
       }
     e->SetDefinedAt(f); 		// Verketten mit Definition/Verwendung
    
     e->Set_nidt(f->Get_nidt());
     qual->Append(e);
     }
   else		// Bezeichner unbekannt
     {
     SimpleError("unknown identifier",e->pos,e->leng);
     e->SetSort(new Sort (s_const));
     return e;
     }
   e = s->GetNextElem();
   }
 if (!e)	// nur Qualifier gefunden
   {
   e = (SymListElem*) s->GetLastElem();
   SimpleError("there are only qualifiers",e->pos,e->leng);
   e->SetSort(new Sort (s_const));
   return e;
   }
 
 UseOfID_help(f,e,s);
 delete qual;
 delete empty;
 return e;
 }

void Environment::AppendUnknown(SymList* s)
 {
 assert(s);
 Frame *p = (Frame*) GetFirstElem();
 assert(p);
 SymListElem *e = s->GetFirstElem();
 assert(e);
 e->SetQualifier(s);
 p->slist.Append(e);
 }

void Environment::AppendAndPopUnknownIDs(int isHeader, Environment *Env)
 {
 Frame *p,*q,*r;
 SymListElem *e,*d;
 SymList *empty = new SymList;
 SymList *s;
 int isglobal;

 assert(Env);
 r = (Frame*) Env->GetFirstElem();	// Top Rahmen von Env
 assert(r);
 p = (Frame*) GetFirstElem();		// Top Rahmen von UnknownIDEnv
 assert(p);
 q = (Frame*) GetNextElem();		// Top umfassender Rahmen von
 assert(q);				// UnknownIDEnv (existiert immer, da
 					// globaler Rahmen nicht gepoppt werden
                                        // darf)
 isglobal = (GetNextElem()==0);		// Ist q globaler Rahmen des
 					// UnknownIDEnv ?
 e = p->slist.GetFirstElem();		// Bezeichner des Top Rahmen durch-
 while (e)				// laufen
   {
   if (isglobal)
     {
     d = Env->Lookup(e,empty);		// Suche in ganz Env
     }
   else
     {
     d = r->slist.Lookup(e,empty);	// Suche in Top Rahmen von Env
     }
   if (d)				// Wenn gefunden
     {
     s = e->GetQualifier();
     UseOfID_help(d,s->GetFirstElem(),s);// Bezeichner wurde gebunden 
     }
   else					// Bezeichner wurde nicht gebunden
     {
     if (isglobal && !isHeader)
       {
       e->Set_nidt(nidt_normal);	// Bezeichner kann nicht mehr gebunden
       SimpleError("undefined symbol",e->pos,e->leng);	// werden
       }
     else				// Bezeichner kann spaeter noch
       {				// gebunden werden.
       q->slist.Append(e);		// Im umgebenden UnknownIDEnv
					// (evtl. gleichen Bezeichner mehrfach)
       }				// eintragen
     }
   e = p->slist.GetNextElem();
   }
 delete empty;
 DelFirstElem();			// Pop UnknownIDEnv
 }

void Environment::PopUnknownIDs()
 {
 Frame *p,*q;
 SymListElem *e;

 p = (Frame*) GetFirstElem();		// Top Rahmen von UnknownIDEnv
 assert(p);
 q = (Frame*) GetNextElem();		// Top umfassender Rahmen von
 assert(q);				// UnknownIDEnv (existiert immer, da
 					// globaler Rahmen nicht gepoppt werden
                                        // darf)
 e = p->slist.GetFirstElem();		// Bezeichner des Top Rahmen durch-
 while (e)				// laufen
   {
   q->slist.Append(e);			// Im umgebenden UnknownIDEnv eintragen
   e = p->slist.GetNextElem();
   }
 DelFirstElem();			// Pop UnknownIDEnv
 }

void Environment::AppendRestUnknownIDs(Environment *Env)
 {
 Frame *p;
 SymListElem *e,*d;
 SymList *s;

 assert(Env);
 p = (Frame*) GetFirstElem();		// Top Rahmen von UnknownIDEnv
 assert(p);
 e = p->slist.GetFirstElem();		// Bezeichner des Top Rahmen durch-
 while (e)				// laufen
   {
   s = e->GetQualifier();
   d=Env->UseOfID(s);			// Bezeichner pruefen
   e = p->slist.GetNextElem();
   }
 }



SymList* Environment::Export(SymList *exportlist)
 {
 Frame  *p = (Frame*) GetFirstElem();
 SymListElem *q, *foundq, *qual, *foundqual;
 SymList     *dummy = new SymList;
 SymList     *result = new SymList;
 normalID_Type  t;

 assert(p);
 if (exportlist->GetLength()==0)		// Wenn alles exportiert
   {
   q = (SymListElem*) p->slist.GetFirstElem(); // Fuer alle Elemente
   while (q)
     {
     t = q->Get_nidt();				// qualifier,
     if ( (t==nidt_qualifier)     ||		// globale Parameter,
	  					// Operatoren,
	  (t==nidt_formalglobPar) ||		// Assumptions und
	  (t==nidt_operator)      ||		// Theorems
	  (t==nidt_fctOperator)   ||		// werden exportiert.
	  (t==nidt_assumption)    ||
	  (t==nidt_theorem)        )
       {
       result->Append(q);
       }
     q = (SymListElem*) p->slist.GetNextElem();// Naechstes Element
     }
   }
 else					// Wenn Exporte explizit genannt
   {
   q = (SymListElem*) p->slist.GetFirstElem();
   while (q)
     {
     t = q->Get_nidt();
     if ( 				// Globale Parameter und
	  (t==nidt_formalglobPar) ||	// Assumptions werden immer
	  (t==nidt_assumption)  )	// exportiert.
       {
       result->Append(q);
       }
     q = (SymListElem*) p->slist.GetNextElem();
     }

					// Suche zu exportierende Bezeichner
					// und ueberpruefe ihre Qualifier
					// Durchlaufe dazu Exportliste:

   q = (SymListElem*) exportlist->GetFirstElem();
   while (q)
     {
     qual = (SymListElem*) q->GetQualifier()->GetFirstElem();
     while (qual)
       {
       foundqual = Lookup(qual,dummy);	// Qualifier in Env bekannt ?
       if (! foundqual)
	 {				// Nein
         SimpleError("unknown qualifier",qual->pos,qual->leng);
         }
       else				// Ja
         {
         qual->Set_nidt(nidt_qualifier);
         qual->SetDefinedAt(foundqual);
         foundqual->SetPrime(1);	// Dann markieren und pruefen, ob
					// Anzahl formale Parameter des
					// Qualifiers korrekt angegeben

	 if (qual->GetQualifier()->GetLength() != // ( Qualifier von qual sind
	     foundqual->GetSort()->GetLength()-1) //   hier die formalen
						  //   Parameter von qual )
           {
           SimpleError("wrong number of formal Parameter",
                       qual->pos,qual->leng);
           }
         }
       qual = (SymListElem*) q->GetQualifier()->GetNextElem();
       } 
     q = (SymListElem*) exportlist->GetNextElem();
     }
     delete dummy;

					// Alle markierten Qualifier exportieren

   q = (SymListElem*) p->slist.GetFirstElem();
   while (q)
     {
     t = q->Get_nidt();
     if ( (t==nidt_qualifier) &&
	  (q->is_primed())  )
       {
       result->Append(q);
       }
     q = (SymListElem*) p->slist.GetNextElem();
     }

					// Zum Schluss alle zu exportierenden
					// Definitionen ausgeben

   q = (SymListElem*) exportlist->GetFirstElem();
   while (q)
     {
     foundq = Lookup(q,q->GetQualifier()); // Operator in Env suchen
     if (! foundq)
       {
       SimpleError("undefined symbol",q->pos,q->leng);
       }
     else
       {
       t = foundq->Get_nidt();		// Wenn gefunden, dann pruefen
       if ( (t==nidt_operator)    ||	// ob Operator oder Theorem 
	    (t==nidt_fctOperator) ||	// vorliegt
	    (t==nidt_theorem)      )
         {
         if (t==nidt_operator)		q->Set_nidt(nidt_expOperator);
         if (t==nidt_fctOperator)       q->Set_nidt(nidt_expOperator);
         if (t==nidt_theorem)		q->Set_nidt(nidt_expTheorem);
         q->SetDefinedAt(foundq);
         q->SetSort(foundq->GetSort()->copy());
         result->Append(foundq);
         }
       else
         {
         SimpleError("only operators or theorems may appear in export list",
                     q->pos,q->leng);
         }
       }
     q = (SymListElem*) exportlist->GetNextElem();
     }
   }
 return result;
 }

Sort_Type Environment::GetDefaultParSort()
 {
 return ((Frame*) GetFirstElem())->defaultParSort;
 }

Sort_Type Environment::GetDefaultOpSort()
 {
 return ((Frame*) GetFirstElem())->defaultOpSort;
 }

Sort_Type Environment::GetDefaultSort()
 {
 return ((Frame*) GetFirstElem())->defaultSort;
 }

void Environment::SetDefaultParSort(Sort_Type defSort)
 {
 ((Frame*) GetFirstElem())->defaultParSort = defSort;
 }

void Environment::SetDefaultOpSort(Sort_Type defSort)
 {
 ((Frame*) GetFirstElem())->defaultOpSort = defSort;
 }

void Environment::SetDefaultSort(Sort_Type defSort)
 {
 ((Frame*) GetFirstElem())->defaultSort = defSort;
 }

Module_Type Environment::GetModuleType()
 {
 return ((Frame*) GetFirstElem())->MType;
 }

void Environment::SetModuleName(SymListElem *e)
 {
 ((Frame*) GetFirstElem())->MName = e;
 }

SymListElem* Environment::GetModuleName()
 {
 return ((Frame*) GetFirstElem())->MName;
 }

int Environment::Recover()
 {
 Frame *p;
 while ( ( p=(Frame*) GetFirstElem() ) && ! p->RecoveryPoint)
   {
   DelFirstElem();
   }
 assert(p);
 return p->MDepth;
 }
