 /* ----------------------------------------------------------------------*
  * HiGraph.c
  *
  * Michael Geikowski, 24.3.93
  *
  * Enthaelt die Member-Funktionen der Klassen HiGraph und LevelInfo zur
  * Verwaltung von Hierarchie-Graphen.
  *----------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
 
#include "higraphgnu.h"

/*----------------------------------------------------------------------*
 * Klasse HiGraph
 *----------------------------------------------------------------------*/

HiGraph::HiGraph(AdjList *al, int NrOfLevels)
 {
 assert(al && NrOfLevels > 0);

 Node* n;
 int   Level;
 char  NextFreePos[NrOfLevels];

 Levels = NrOfLevels;
 the_Graph = new LevelInfo[NrOfLevels];

 n = al->GetFirstNode();	// Knoten je Level zaehlen und durchnummerieren
 while (n)
   {
   Level = n->GetLevel();
   assert(0<=Level && Level < NrOfLevels);
   the_Graph[Level].NrOfNodes++;
   n->SetPos(the_Graph[Level].NrOfNodes);
   n->SetTempPos(the_Graph[Level].NrOfNodes);
   n = al->GetNextNode();
   }
 for (Level=0; Level < NrOfLevels; Level++)	// KnotenArray und Queue je
   {						// Level anlegen
   NextFreePos[Level]=0;
   the_Graph[Level].NodeArray = new Node* [the_Graph[Level].NrOfNodes];
   the_Graph[Level].queue = new PQueue (the_Graph[Level].NrOfNodes);
   
   }
 n = al->GetFirstNode();	// Knoten in KnotenArrays einordnen
 while (n)
   {
   Level = n->GetLevel();
   the_Graph[Level].NodeArray[NextFreePos[Level]]=n;
   NextFreePos[Level]++;
   n = al->GetNextNode();
   }
 }

HiGraph::~HiGraph()
 {
 assert(the_Graph);
 delete [ Levels ] the_Graph;			// LevelInfos loeschen
 }

Node* HiGraph::GetNodeOfLevel(int Level, int nr)
 {
 assert(the_Graph && 0<=Level && Level<Levels);
 LevelInfo *l = &the_Graph[Level]; 
 assert(l->NodeArray && 0<=nr && nr < l->NrOfNodes);
 return(l->NodeArray[nr]);
 }

void HiGraph::SetNodeOfLevel(Node *n, int Level, int nr)
 {
 assert(the_Graph && 0<=Level && Level<Levels);
 LevelInfo *l = &the_Graph[Level]; 
 assert(l->NodeArray && 0<=nr && nr < l->NrOfNodes);
 l->NodeArray[nr] = n; 
 }

Node* HiGraph::FirstNodeOfLevel(int Level)
 {
 assert(the_Graph && 0<=Level && Level<Levels);
 LevelInfo *l = &the_Graph[Level]; 
 assert(l->NodeArray);
 l->current=1;			// Durchlaufzaehler initialisieren
 return(l->NodeArray[0]);
 }

Node* HiGraph::NextNodeOfLevel(int Level)
 {
 assert(the_Graph && Level<Levels);
 LevelInfo *l = &the_Graph[Level]; 
 assert(l->NodeArray);
 if (l->current < l->NrOfNodes) // Durchlauf beendet ?
   {
   l->current++;		// nein => naechstes Element
   return(l->NodeArray[l->current-1]);
   }
 return 0;			// ja => NULL
 }

/*----------------------------------------------------------------------*
 * Klasse LevelInfo
 *----------------------------------------------------------------------*/

LevelInfo::LevelInfo()
 {
 NrOfNodes = 0;
 NodeArray = 0;
 current   = 0;
 queue     = 0;
 }

LevelInfo::~LevelInfo()
 {
 if (NodeArray) delete [ NrOfNodes ] NodeArray;
 if (queue) delete queue;
 }
