/*----------------------------------------------------------------------*
 * List.c
 *
 * Michael Geikowski, 21.4.93
 *
 * Enthaelt die Member-Funktionen der Klassen ListLink und List.
 *----------------------------------------------------------------------*/

#include <assert.h>

#include "list.h"

/*----------------------------------------------------------------------*
 * Klasse ListLink
 *----------------------------------------------------------------------*/

ListLink::ListLink(ListElem* e)
 {
 assert(e);
 elem = e;
 next = 0;
 elem->count+=1;		// Element-counter erhoehen.
 }

ListLink::~ListLink()
 { 
 assert(elem);
 elem->count-=1;		// Element-counter runterzaehlen
 if (elem->count==0)		// Nur, wenn Element in keiner anderen
   {				// Liste mehr auftritt, kann es
   delete elem;			// geloescht werden.
   }
 }

/*----------------------------------------------------------------------*
 * Klasse List
 *----------------------------------------------------------------------*/

List::List(ListElem *e)
 {
 ListLink *ll = 0;

 length = 0;
 if (e!=0)			// Wenn nicht leere Liste.
   {
   ll = new ListLink(e);	// einelementig
   length++;
   }
 first = last = ll;
 current = 0;			// Durchlauf-Merker
 }

List::~List()
 {
 ListLink *p,*q;

 p=first;			// Liste loeschen heisst alle Links
 while (p!=0)			// loeschen. Links loeschen evtl. Elemente.
  {
  q = p->next;
  delete p;
  p = q;
  }
 }

void List::Append(List *l)
 {
 assert(l);
 if (l->first != 0)		// Wenn l nicht leer...
   {
   if (first==0)		// Wenn this leer...
     {
     first = l->first;		// this = l
     last  = l->last;
     }
   else
     {
     last->next = l->first;	// l an this anhaengen
     last       = l->last;
     }
   }
 length += l->length;		// Laenge aktualisieren
 l->first = l->last = 0;	// l loeschen. Links von l gehoeren nun zu this
 delete l;
 }

void List::Append(ListElem *e)
 {
 ListLink *ll = new ListLink(e);

 assert(e);
 if (first==0)			// Wenn this leer...
   {
   first = last = ll;
   }
 else
   {
   last->next = ll;		// ll hinten anhaengen
   last       = ll;
   }
 length++;
 }

ListElem* List::GetFirstElem()
 {
 if (first)
   {
   current = first;		// Liste nicht leer.
   return first->elem;
   }
 else
   {
   return 0;			// Liste leer.
   }
 }

ListElem* List::GetLastElem()
 {
 assert(last);
 return last->elem;		// Liste darf nicht leer sein.
 }

ListElem* List::GetNextElem()
 {
 if (current)			// Wenn Durchlauf gestartet wurde.
   {
   current = current->next;
   if (current)			// Ende noch nicht erreicht.
     {   
     return current->elem;
     }
   }
 return 0;			// Durchlauf beendet.
 }

ListElem* List::GetCurrentElem()
 {
 if (current)
   {
   return current->elem;
   }
 return 0;
 }

void List::SetCurrentElem(ListElem *e)
 {
 assert(current && e);
 current->elem->count-=1;	// aktuelles Element in einer Liste weniger
 if (current->elem->count==0)	// evtl. loeschen
   {
   delete current->elem;
   }
 e->count+=1;			// e in einer Liste mehr
 current->elem = e;
 }

void List::DelCurrentElem()
 {
 assert(current);
 if (current!=first)		// evtl. erstes Element fuer aktuelles einsetzen
   {
   SetCurrentElem(first->elem);
   }
 DelFirstElem();		// erstes Element loeschen
 }

void List::DelFirstElem()
 {
 ListLink *p;

 assert(first);
 p=first->next;			// Zweites Element.
 delete first;
 if (p==0)
   {
   first=last=0;		// Liste jetzt leer.
   }
 else
   {
   first=p;			// Altes zweites ist neues erstes Element
   }
 length--;
 }

void List::NewFirstElem(ListElem *e)
 {
 ListLink *ll = new ListLink(e);

 assert(e);
 if (first==0)			// Wenn this leer...
   {
   first = last = ll;
   }
 else
   {
   ll->next = first;		// ll vorne einhaengen
   first    = ll;
   }
 length++;
 }

int List::is_Singleton()
 {
 return (length == 1);
 }

int List::GetLength()
 {
 return length;
 }

void List::Set_current_Link(ListLink* li)
 { 
  current = li;
 }

ListLink* List::Get_current_Link()
 {
   return current;
 }
