/*-------------------------------------------------------------------------
 * Semconcept.h
 *
 * Olaf Meier,  10.9.94
 *
 * Dieses Modul bereitet die zur semantischen Konzepterfassung notwendigen Informationen
 * auf. U.a. ist es fuer die Expansion der Systemformel zustaendig. Des weiteren werden
 * die in der Systemformel auftretenden Aktionen ermittelt. Diese Aktionen mit den jeweils
 * genutzten Systemvariablen werden abschliessend in eine Systemvariablen-Datei zwischen-
 * gespeichert, damit sie spaeter von der Browserkomponente gelesen werden koennen.
 * 
 *--------------------------------------------------------------------------------------------*/ 

#ifndef SEMCONCEPT_H
#define SEMCONCEPT_H

#include "list.h"
#include "symlist.h"
#include "adjlistgnu.h"
#include "parsetree.h"


// Flags zur Untersuchung der kanonischen Systemformel.
// Die Formel bsitzt zwei Stufen:
// 1.Stufe : ODER - Verknuepfung der verschiedenen Aktionen
// 2.Stufe : UND - Verknuepfung der Praedikate in den jeweiligen Aktionen
static enum Node_Level {AND_Level, OR_Level};

// Die Hilfs-Klasse IDInfo verwaltet die jeweiligen formalen Parameter
// eines Bezeichners, sowie ihre aktuellen Argumente

class IDinfo : public ListElem
 {
   public : 

   IDinfo(): parname(0), expression(0) {};
   ~IDinfo(){};

   char* GetParname()
    { return parname;}

   void SetParname(char* n)
    { parname = n;}
  
   ParseExpr* GetExpression()
    { return expression ;}

   void SetExpression(ParseExpr* pe)
    { expression = pe ;}

   private :

    char* parname;  // Name des lokalen Parameters
    ParseExpr* expression;  // Aktuelles Argument des Parameters
 };

// Die Hilfs-Klasse QualInfo verwaltet die jeweiligen Qualifier innerhalb der Spezifikation
// Zusaetzlich werden evtl. die Namen der Modulparameter mitverwaltet.
// Bei der Suche nach Bezeichnerdefinitionen innerhalb des Ableitungsbaums
// muessen die Qualifier des angewendeten Bezeichners 
// beachtet werden. Hierbei haengt die Wahl des entsprechenden Teilbaums bei der Suche
// nach Bezeichnerdefinitionen vom jeweiligen Qualifier ab. 'passed' gibt 
// an, ob der durch den Qualifier bestimmte Teilbaum durchsucht wurde oder nicht.  
// Die richtige Bezeichnerdefinition ist gefunden, falls die
// Teilbaeume alle Qualifier des jeweiligen Bezeichners abgearbeitet wurden.
 

class QualInfo : public ListElem
 {
    public :

    QualInfo(char* qname)
     : qualname(qname) , passed(0)
     { 
       parlist =  new List;    
      };

    ~QualInfo()
     {
       delete parlist;
     };

    char* GetQualName()
     { return qualname;};


    void SetPassed()
     { passed = 1;}

    int isPassed()
     {return (passed == 1);}

    List* GetQualParams()
     {
       
         return parlist;
       
     };
    private :
       int passed;   // ...wurde entsprechender Teilbaum betreten ? 
       char* qualname;  // Name des Qualifiers
       List* parlist;  // Modulparameterliste des Qualifiers
 };

// Bei einigen der unten aufgefuehrten Routinen taucht desoefteren der 
// Parameter 'start' auf. Routinen mit diesem Parameter werden rekursiv aufgeruefen
// und nutzen diesen Parameter zur effizienteren Standortbestimmung innerhalb
// des Ableitungsbaums. 



// Die Routine GetSubstitutions()  ermittelt die durchzufuehrenden Parametersubstitutionen innerhalb
// des includierten Moduls, die 'substlist' eingetragen werden. 'expr' ist
// ein  Verweis auf den aktuellen Baumteil. 'start' gibt den jeweiligen Startknoten
// der rekursiv aufgerufenen Routine an. 'sl' enthaelt eine Liste kopierter
// Symbole der Symboltabelle. Diese wird aus implementierungstechnischen
// Gruenden zum spaeteren Loeschen der Symbole benoetigt. 

void GetSubstitutions(ParseExpr* expr,ParseExpr* start, List* substlist,SymList* sl);

// Hilfroutine CollectSubstitutions() fuer GetSubstitutions() zur Ermittlung der Substitutionen, die
// in einer Liste zurueckgegenen werden.

List* CollectSubstitutions(ParseExpr* expr,ParseExpr* start,SymList* sl);


// Die Hilfsroutine ParameterSubstitution()  ist fuer den eigentlichen Austausch
// von Bezeichnern durch andere Ausdruecke innerhalb des Ableitungsbaums zustaendig
 
void ParameterSubstitution(ParseExpr* expr, List* substlist,SymList* sl);

// Routine TreeImport()  verkettet den kompletten Ableitungsbaum des includierten bzw.
// importierten Moduls  mit dem Ableitungsbaum des aktuellen
// Moduls.

void TreeImport(ParseExpr* tree,ParseExpr* start,SymList* sl);

// Die Routine CopyTree()  kopiert den Ableitungsbaum 'expr' und gibt die Kopie zurueck.
// Kopierte Terminalsymbole werden der Symbolliste 'sl' zugefuegt.

ParseExpr* CopyTree(ParseExpr* expr,SymList* sl);

// Hilfsroutine CorrectTree() , die Recordkomponentennamen vom eigentlichen benutzerdefinierten Bezeichner
// (ID) trennt, um die Informationsaufbereitung fuer die semantische Konzepterfassung
// zu erleichtern.

void CorrectTree(ParseExpr* expr);

// Die Routine  ExamineSystemFormula()  sucht nach der Systemformel innerhalb der Spezifikation und expandiert diese.
// 'expr' und 'start' dienen der Orientierung bei der Traversierung des Ableitungsbaumes.
// 'sl' enthaelt die Liste kopierter Terminalsymbole.

void  ExamineSystemFormula(ParseExpr* tree,ParseExpr* start,SymList* sl);  //Bei ersten Aufruf: tree == start

// ExpandExpression(), Hilsroutine von ExamineSystemFormula(), die einen angewendeten Bezeichner
//  im Ableitungsbaum durch
// dessen Definition ersetzt.

void  ExpandExpression(ParseExpr* expr,ParseExpr* start,SymList* sl);

// Routine CollectQualifers() zur Ermittlung aller Qualifizierer eines Bezeichners, die in 'qualList' vermerkt werden.
 
void CollectQualifers(ParseExpr* expr,List* qualList,ParseExpr* start,SymList* sl);

// Routine CollectActualParams()  zur Ermittlung der aktuellen Argumente eins parametrisierten Bezeichners.
// Die Argumente werden in 'parList' gespeichert.
void CollectActualParams(ParseExpr* expr,List* parList);

// Routine SearchforOperatorDef() fuer das Suchen der Definition eines Bezeichners 'opname' mit der 
// Qualifizierliste 'qualList' und der aktuellen Argumentliste 'parList' im 
// gesamten Ableitungsbaum der Spezifikation. Ein Verweis auf den Teilbaum einer Bezeichnerdefinition
// wird zurueckgeliefert.

ParseExpr* SearchforOperatorDef(ParseExpr* expr,ParseExpr* start,char* opname, List* qualList,List* parList,SymList* sl);

// Die Hilfsroutine  Examine_SymbolDef_LHS() untersucht eine Bezeichnerdefinition daraufhin, ob es sich
// um den Bezeichnername 'opnamen' handelt und ob alle Qualifizierer der Liste 'qualList'
// gefunden wurden. Im Erfolgsfall (Rueckgabewert == 1) werden evtl. formale Parameter des Bezeichners ermittelt und 
// durch die aktuellen Argumente aus Liste 'parList' in der Definition ersetzt. 
int Examine_SymbolDef_LHS(ParseExpr* expr, char* opname, List* parList,List* qualList);

// CollectFormalParameters_formalPar_ssl(),Hilsroutine fuer SearchforOperatorDef() zur 
// Ermittlung der formalen Parameter des gesuchten Bezeichners.
// Die Parameter werden in der Liste 'parList' vermerkt.
void CollectFormalParameters_formalPar_ssl(ParseExpr* expr,List* parList);

// CollectFormalParameters_formalPar(), Hilfroutine fuer SearchforOperatorDef() zur Ermittlung 
// der formalen Parameter des gesuchten Bezeichners.
// Die Parameter werden in der Liste 'parList' vermerkt.
void CollectFormalParameters_formalPar(ParseExpr* expr, List* parList);

// CollectFormalParameters_normalID_csl(), Hilfsroutine fuer SearchforOperatorDef() zur
// Ermittlung der formalen Parameter des gesuchten Bezeichners.
// Die Parameter werden in der Liste 'parList' vermerkt.
void CollectFormalParameters_normalID_csl(ParseExpr* expr, List* parList);


// Die Routine isUNCHANGEDconstruct()  prueft, ob der Teilbaum 'expr' den Aufbau <ausdruck' = ausdruck> (UNCHANGED-Konstrukt)
// Im Erfolgsfall wird 1 , sonst 0 zurueckgeliefert.
int isUNCHANGEDconstruct(ParseExpr* expr);

// CompareExprs(), Hilfsroutine  fuer isUNCHANGEDconstruct(), die ueberrueft ob die 
// Teilbaeume 'expr1' und 'expr2' identisch
// sind. Im Erfolgsfall wird 1 , sonst 0 zurueckgeliefert.
int CompareExprs(ParseExpr* expr1, ParseExpr* expr2);

// Die Routine ExamineTerm()  untersucht das Praedikat, welches durch den Teilbaum 'expr' dargestellt wird.
// 'actionnr' gibt hierbei die Aktionskennung, in der sich das Praedikat befindet an. 
void ExamineTerm(ParseExpr* expr, int actionnr);

// VariableAccessSearch() durchsucht das durch den Teilbaum 'expr' dargestellte Praedikat
// auf Systemvariablen und evtl. vorhanden Komponenten. Wird die Anwendung
// einer Systemvariablen gefunden, so ist 'foundvar' gleich 1, sonst 0. 
// 'openbraces' dient der Klammer-Kontrolle von geschachtelten Systemvariablen-Ausdruecken.  
void VariableAccessSearch(ParseExpr* expr,int* foundvar,int* openbraces);


// Hauproutine CreateSemanticConcept()  zur Untersuchung der Systemformel. 'varexpr' verweist
// auf die Liste  aller in der Systemformel benutzten Systemvariablen, 'actionexpr' verweist
// auf die eigentliche kanonische Systemformel.
// Jede in der Systemformel angewendete Variable wird mit ihren Aktionskennungen in der 
// Adjazenzliste 'SCadjl' vermerkt.
void CreateSemanticConcept(ParseExpr* varexpr,ParseExpr* actionexpr,AdjList* SCadjl);

// Die Routine ExamineOperation()  traversiert rekursiv den Teilbaum, der die Systemformel darstellt.
// 'SCadjl' ist die aktuelle Liste gefundener Sytemvariablen(-komponenten). 'actionnr' 
// gibt die aktuelle Aktion  an, in der sich die Routinen gerade befindet.
// Die aktuelle Aktionsnummer 'actionnr'  wird jeweils in 
// Abhaengigkeit der augenblicklich erreichten Stufe (Disjunktions bzw. Konjunktions-Ebene) 
// innerhalb der kanonischen Systemformel gewaehlt (beim Uebergang von der OR-Ebene in 
// eine AND-Ebene wird actionnr um 1 erhoeht)
 
void ExamineOperation(ParseExpr* expr,AdjList* SCadjl, Node_Level nl, int actionnr);

// Routine AppendVariablesToAdjList() fuegt die in der Systemformel gefundenen Systemvariablen bzw. ihre
// Komponenten unter Angabe der jeweiligen Aktionsnummer 'actnr' in die Adjazenzliste 'SCadjl'
// ein.
void AppendVariablesToAdjList(AdjList* SCadjl,int actnr);


#endif
