/*----------------------------------------------------------------------*
 * Sort.h
 *
 * Michael Geikowski, 5.6.93
 *
 * Definiert die Klassen Sort und SortListElem sowie einige mit der
 * Klasse Sort verwandte Non-Member-Funktionen zur Unterstuetzung der
 * Sortenueberpruefung durch den Parser.
 *----------------------------------------------------------------------*/

#ifndef SORT_H
#define SORT_H

#include <stdio.h>
#include <assert.h>

#include "list.h"

// Flags fuer Sorten ( s_unknown fuer Fehler, s_nested fuer zusammengesetzte
// Sorten, s_joker passt auf alles, Rest fuer primitive Sorten )

enum Sort_Type
 { s_action   = 'A', s_bool     = 'B', s_const    = 'C', s_pred = 'P',
   s_transfct = 'R', s_statefct = 'S', s_temporal = 'T',
   s_joker    = 'J', s_unknown  = 'X', s_nested   = 'Z'};

class Sort;

class SortListElem : public ListElem
//==================================
	// Ein Objekt der Klasse Sort ist eine Liste von Objekten der
        // Klasse SortListElem.
 {
 friend class Sort;

 public:

   SortListElem(Sort_Type t, Sort *s = 0)
 //--------------------------------------
	:the_Sort(t), nested(s)
	{ assert( ((s==0) && (t!=s_nested)) ||
                  ((s!=0) && (t==s_nested)) ); }

   ~SortListElem();
 //----------------
	// loescht gegebenenfalls 'nested'

   Sort_Type GetSort()
 //-------------------
	{ return the_Sort; }

   void SetSort( Sort_Type t)
 //--------------------------
	{ the_Sort = t; }

 private:

   Sort_Type the_Sort;	// Flag
   Sort      *nested;	// Zeiger auf verschachtelte Sorte (nur bei Flag
   			// gleich `nested' verwendet)
 };

class Sort : public List
//======================
	// Die Klasse Sort stellt TLA+ Sorten als Listen von Objekten der
        // Klasse SortListElem dar.
	// Eine Sorte ist entweder primitiv oder zusammengesetzt. Primitive
        // Sorten werden dargestellt als einelementige Liste mit
        // SortListElem::the_Sort != nested.
        // Zusammengesetzte Sorten werden durch beliebig lange Listen 
        // beschrieben. Die Elemente einer zusammengesetzten Sorte
        // koennen primitiv (SortListElem::the_Sort != nested) oder
        // zusammengesetzt sein (SortListElem::the_Sort == nested). Im letzten
        // Fall zeigt SortListElem::nested auf die zusammengesetzte Sorte.
 {
 public:

   Sort(Sort_Type t)		// einelementige Liste mit Sorten-Flag t
 //-----------------
	{ List::Append(new SortListElem(t)); }


   Sort(SortListElem *e=0)	// ein- oder nullelementige Liste
 //-----------------------
	: List(e)
	{}

   ~Sort()
 //-------
	{}

   void Append(Sort_Type t)	// Verlaengere Liste um ein Element mit
   				// Sorten-Flag t
 //------------------------
	{ List::Append(new SortListElem(t)); }

   int operator==(Sort_Type t);
 //----------------------------
	// Ergibt True, wenn this eine primitive Sorte beschreibt, die durch
        // das Sorten-Flag t gegeben ist.
        // Ergibt False sonst.

   int operator<=(Sort_Type t);
 //----------------------------
	// Ergibt True, wenn this eine primitive Sorte beschreibt, die gemaess
        // der Halbordnung der primitiven Sorten kleiner oder gleich der
        // primitiven Sorte ist, die durch das Sorten-Flag t gegeben wird.
	// Ergibt False sonst.
        
   int operator<=(Sort& s);
 //------------------------
	// Ergibt True, wenn this die durch s beschriebene Sorte matched.
        // Ergibt False sonst.
        // Eine Sorte s1, die durch die Liste l1 beschrieben ist, soll eine
        // Sorte s2, die durch die Liste l2 beschrieben ist, matchen, wenn
        // beide Listen gleiche Laenge haben und fuer jedes Element e von l1
        // gilt: Wenn e das i-te Element von l1 ist und e beschreibt eine
        // primitive Sorte, so beschreibt auch das i-te Element f von l2 eine
        // primitive Sorte und e ist kleiner oder gleich f;
        // Wenn e das i-te Element von l1 ist und e beschreibt eine
        // zusammengesetzte Sorte s3, so beschreibt auch das i-te Element f von
        // l2 eine zusammengesetzte Sorte s4 und s3 matched s4.

   int legal_call(Sort& s);
 //------------------------
	// Ergibt True, wenn ein Operator mit Sorte this mit einer aktuellen
        // Parameterliste der Sorte s aufgerufen werden darf, d.h. wenn die
        // Sorte, die aus this gebildet wird, indem das letzte Element der this
        // beschreibenden Liste entfernt wird, die Sorte s matched.
        // Ergibt False sonst.

   Sort* copy();
 //-------------
	// Ergibt eine exakte Kopie von this.

   Sort* raise(int level);
 //-----------------------
        // Hebt die zusammengesetzte Sorte 'this' an in die Stufe 'level'
	// (= 1,2 oder 3) und gibt die angehobene Sorte aus.
        // Ergibt NULL, wenn 'this' nicht zusammengesetzte Sorte oder nicht
        // angehoben werden kann.
 
   void ErrorOutPut(int isParList = 0);
 //------------------------------------
	// 'this' als ausfuehrlichen lesbaren ascii-Text in Error-Datei
 	// schreiben. 'this' wird als Sorte eines Operators ausgegeben, wenn
        // isParList == 0, sonst als Folge der Sorten einer Parameterliste.

   void Sort::ErrorOutPut2();
 //--------------------------
	// 'this' als ausfuehrlichen lesbaren ascii-Text in Error-Datei
	// schreiben. 'this' wird als Sorte eines Operators aufgefasst, von
	// der allerdings nur die Folge der Sorten der Parameter ausgegeben
	// werden.

   SortListElem* GetFirstElem()
 //----------------------------
	{ return (SortListElem*) List::GetFirstElem(); }

   SortListElem* GetLastElem()
 //---------------------------
	{ return (SortListElem*) List::GetLastElem(); }

 };

/*----------------------------------------------------------------------*
 * Non-Member-Funktionen
 *----------------------------------------------------------------------*/

	// Im folgenden sei eine Sorte s1 Untersorte einer primitiven Sorte s2,
        // wenn s1 primitiv ist und gemaess der Halbordnung der primitiven
        // Sorten kleiner oder gleich s2 ist.
	//
        // Korrespondierende Sorten sollen sein:
        //
        // CONSTANT            und BOOLEAN
        // STATE FUNCTION      und PREDICATE
        // TRANSITION FUNCTION und ACTION

Sort* BinaryMaxMatch(Sort *s1, Sort *s2, Sort_Type t,
                     long errpos, int errleng, Sort_Type def);
//------------------------------------------------------------
	// Testet, ob s1 und s2 Untersorten von t sind (t ist Sorten-Flag einer
        // primitiven Sorte). Wenn ja, wird von s1 und s2 die groessere (bzw.
        // eine beliebige, wenn beide gleich gross sind) zurueckgegeben und
        // die andere geloescht. Sonst wird eine Fehlermeldung erzeugt
        // (errpos und errleng geben den Fehlerort an), s1 und s2 werden
        // geloescht und die default Sorte mit Sorten-Flag def wird
        // zurueckgegeben.

Sort* UnaryMaxMatch(Sort *s, Sort_Type t,
                    long errpos, int errleng, Sort_Type def);
//-----------------------------------------------------------
	// Testet, ob s Untersorte von t ist (t ist Sorten-Flag einer
        // primitiven Sorte). Wenn ja, wird s zurueckgegeben. Sonst wird eine
        // Fehlermeldung erzeugt (errpos und errleng geben den Fehlerort an),
        // s wird geloescht und die default Sorte mit Sorten-Flag def wird
        // zurueckgegeben.

void UnaryExactMatch(Sort *s, Sort_Type t, long errpos, int errleng);
//-------------------------------------------------------------------
	// Testet, ob s von Sorte t ist (t ist Sorten-Flag einer
        // primitiven Sorte). Wenn nicht, wird eine Fehlermeldung
	// erzeugt (errpos und errleng geben den Fehlerort an).

Sort* BinaryMaxBoolMatch(Sort *s1, Sort *s2, long errpos, int errleng);
//---------------------------------------------------------------------
	// Testet, ob s1 und s2 Untersorten von TRANSITION FUNCTION sind.
        // Wenn ja, wird die zur groesseren von beiden (oder zu beiden bei
        // Gleichheit) korrespondierende Untersorte von TEMPORAL
        // zurueckgegeben und s1 und s2 werden geloescht. Sonst wird eine
        // Fehlermeldung erzeugt (errpos und errleng geben den Fehlerort an),
        // s1 und s2 werden geloescht und eine Sorte mit Sorten-Flag s_bool
        // wird zurueckgegeben.

Sort* BoolConstToConst(Sort *boolSort, Sort *constSort);
//------------------------------------------------------
	// Wandelt die primitive Untersorte von TEMPORAL `boolSort' in die
        // korrespondierende Untersorte von TRANS FCN. Diese wird mit der
        // primitiven Untersorte von TRANS FCN `constSort' verglichen. Die
        // groessere der beiden Sorten (eine beliebige bei Gleichheit) wird
        // zurueckgegeben. boolSort und constSort werden geloescht.

Sort* BoolConstToBool(Sort *boolSort, Sort *constSort);
//-----------------------------------------------------
	// Ergibt boolSort und loescht constSort, wenn boolSort primitive
        // Sorte mit Sorten-Flag s_temporal ist.
        // Ansonsten:
	// Wandelt die primitive Untersorte von TRANS FCN `constSort' in die
        // korrespondierende Untersorte von TEMPORAL. Diese wird mit der
        // primitiven Untersorte von TEMPORAL `boolSort' verglichen. Die
        // groessere der beiden Sorten (eine beliebige bei Gleichheit) wird
        // zurueckgegeben. boolSort und constSort werden geloescht.

Sort* MaxMatch(Sort *s1, Sort *s2);
//---------------------------------
	// Gibt die Groessere der Sorten s1 und s2 aus und loescht die andere.
        // Beide sind entweder Untersorten von TEMPORAL oder TRANS FCN.

Sort* MaxBoolMatch(Sort *s1, Sort *s2);
//-------------------------------------
	// Berechnet die Groessere der Sorten s1 und s2 (beide sind Untersorten
        // von TRANS FCN). Ausgegeben wird die korrespondierende
	// Untersorte von TEMPORAL. s1 und s2 werden geloescht.

Sort* MaxConstMatchOfMixedSorts(Sort *boolSort, Sort *constSort);
//---------------------------------------------------------------
	// Berechnet die zur Untersorte von TEMPORAL `boolSort'
        // korespondierende Sorte. Diese wird mit der Untersorte von TRANS FCN
        // `constSort' verglichen. Das Maximum wird ausgegeben. boolSort
        // und constSort werden geloescht.

Sort* MaxBoolMatchOfMixedSorts(Sort *boolSort, Sort *constSort);
//--------------------------------------------------------------
	// Ergibt boolSort und loescht constSort, wenn boolSort primitive
        // Sorte mit Sorten-Flag s_temporal ist.
        // Ansonsten:
	// Berechnet die zur Untersorte von TRANS FCN `constSort' 
        // korespondierende Sorte. Diese wird mit der Untersorte von TEMPORAL
        // `boolSort' verglichen. Das Maximum wird ausgegeben. boolSort
        // und constSort werden geloescht.

#endif
