/*-------------------------------------------------------------------------
 * Typeckeck.C
 *
 * Olaf Meier,  10.5.94
 *
 * Dieses Modul uebernimmt die Typpruefung in den einzelnen Modulen der Spezifikation.
 * Die Hauptroutine TypeChecker() startet die Pruefung auf Grundlage des jeweiligen
 * Ableitungsbaums.
 * Die Namen der benutzten Routinen sind an den jeweiligen Produktionsregeln der
 * tTLA+ Grammatik angelehnt. Der Ableitungsbaum wird durch die Routinen traversiert.
 *-------------------------------------------------------------------------------------*/

#include<assert.h>
#include<stdio.h>
#include<string.h>
#include <stdlib.h>
#include "typecheck.h"
#include "dbinterface.h"
#include "parser2.tab.h"
#include "list.h"



// Modulglobale Variablen
int isglobalParDef = 0,         // Typepruefung befindet sich gerade
                                // im Abschnitt PARAMETERS (== 1) 
    isModuleParameter = 0;      // Definition von Modulparametern (==1)

 

     

void MakeActualTypeInfos(SymList* list)
 { TypeComp *tmptype,*dummy;
   ParseExpr* pe;
   SymListElem* se;
   se = list->GetFirstElem();
   while(se)
   { 
     if(se->GetTypeInfo() != NULL)
       { tmptype = se->GetTypeInfo();
         if(tmptype->GetType() == ID_type)
          {
            pe = (ParseExpr*) tmptype->GetID_Type();
            dummy = AppendType_formalPar(pe);  // entsprechender Teilbaum wird nachtraeglich
                                               // ausgewertet.
          };
       };
      se = list->GetNextElem();
   };
 }


int CompareTypeChain(TypeComp* typ1, TypeComp* typ2, int isElemOf = 0)
{ assert(typ1 && typ2);
 
  int ok = 1;
 

  char *name1 = NULL, *name2 = NULL;
  while((typ1 !=NULL) && (typ2 != NULL))
  {  name1 = typ1->GetCompname();
     name2 = typ2->GetCompname();
     if((name1 != NULL) && (name2 != NULL))   // Record-Komponenten sind vorhanden
      ok = (strcmp(name1, name2) == 0);

     ok = ok && CompareTypes(typ1 , typ2,isElemOf);
         

    typ1 = typ1->GetNextType();

    typ2 = typ2->GetNextType();
  }

  if((typ1 != NULL) || (typ2 != NULL))
   return 0;
  else
   return ok;

 }

int CompareTypes(TypeComp* typ1, TypeComp* typ2, int isElemOf = 0)
{ 
  TypeComp *typ3,*typ4;
  TLA_Type t1, t2;

  int ok = 0;

  if((typ1 == NULL) ||(typ2 == NULL))
   return 0;
  else
  { t1 = typ1->GetType();
    t2 = typ2->GetType();
    if((t1==UNKNOWN_type) || (t2 == UNKNOWN_type) )
      return 0;
    else if((t1== ANY_type)||(t2 == ANY_type))
      return 1;
    else if(
            (((t1== SET_type)||(t1 ==SUBSET_type)||(t1==ENUMSET_type)) && (t2== EMPTYSET_type)) ||
            ((t1 == EMPTYSET_type) && ((t2 == SUBSET_type)||(t2 ==SET_type)||(t2==ENUMSET_type))) ||
            (( t1 == EMPTYSET_type) && (t2 == EMPTYSET_type))
           )
     return 1;
    else if((t1 == UNION_type)&&(t2!=UNION_type))
     {
      typ3 = typ1->GetComp1();

      ok = CompareTypes(typ3,typ2,isElemOf);
      typ3 = typ3->GetNextType();
      while(typ3 != NULL)
      {
       if(isElemOf)
        ok = (ok && CompareTypes(typ3, typ2,isElemOf));
       else
        ok = (ok || CompareTypes(typ3,typ2,isElemOf));
       typ3= typ3->GetNextType();
      };
      return ok;
     }
    else if((t1!= UNION_type)&&(t2 == UNION_type))
     {
       typ3 = typ2->GetComp1();
       ok = CompareTypes(typ1,typ3,isElemOf);
       typ3 = typ3->GetNextType();
       while(typ3 != NULL)
       {
        ok = (ok || CompareTypes(typ1, typ3,isElemOf));
        typ3 = typ3->GetNextType();
       };
       return ok;
     }
    else if((t1 == UNION_type)&&(t2 == UNION_type))
    { typ3 = typ1->GetComp1();

       while(typ3 != NULL)
        {
         typ4 = typ2->GetComp1();
        while(typ4 != NULL)
         {
           ok = (ok || CompareTypes(typ3,typ4,isElemOf));
           typ4 = typ4->GetNextType();
         };
         typ3 = typ3->GetNextType();
       };
       return ok;
     }

    else if(
            ((t1==SET_type)&& (t2 == SET_type)) ||
            ((t1==SUBSET_type) && (t2 == SET_type)) ||
            ((t1==SET_type) && (t2 == SUBSET_type)) ||
            ((t1==SUBSET_type) && (t2==SUBSET_type))
            )

     {
      typ3 = GetElementofType(typ1);
      typ4 = GetElementofType(typ2);
      ok = CompareTypes(typ3,typ4,isElemOf);
      delete typ3;
      delete typ4;
      return ok;
     }
    else if(
            ((t1==SET_type)&& (t2 ==ENUMSET_type)) ||
            ((t1==ENUMSET_type) && (t2 == SET_type)) ||
            ((t1==SET_type) && (t2 == ENUMSET_type)) ||
            ((t1==ENUMSET_type) && (t2==ENUMSET_type))
            )

     {
      typ3 = GetElementofType(typ1);
      typ4 = GetElementofType(typ2);
      ok = CompareTypes(typ3,typ4,isElemOf);
      delete typ3;
      delete typ4;
      return ok;
     }
    else if(((t1== RECORD_type)|| (t1 == TUPLE_type)) && (t1==t2) )
    {
      typ3 = typ1->GetComp1();
     typ4 = typ2->GetComp1();

      assert(typ3 && typ4);
       ok = CompareTypeChain(typ3,typ4,isElemOf);

       return ok;

     }
   else if((t1== ARRAY_type)&&(t1 == t2))
     {
       typ3 = typ1->GetComp1();
       typ4 = typ2->GetComp1();
       assert(typ3 && typ4);
       ok = CompareTypeChain(typ3,typ4,isElemOf);


       typ3 =  typ1->GetComp2();
       typ4 =  typ2->GetComp2();
       assert(typ3 && typ4);
      ok = (ok && CompareTypeChain( typ3,typ4, isElemOf));
       return ok;
     }

   else
    {
     ok = (t1==t2);
     return ok;
    }

  } // Ende von aesserem else
} // Ende von CompareTypes


     


// Hauptroutine, die nach dem Pass2-Durchgang aufgerufen wird
void TypeChecker(ParseExpr* tree)  
 { assert(tree);
  
   ParseExpr *actexpr,*pe;
   int index,max;
   noTerminal_Type nt,tmpnt;
   Symbol_Type st;
   max = tree->GetMaxIndex();
   ParseSymb *actps;
   SymListElem* se;
   TypeComp *dummy;
   for(index=1;index <= max+1; index++)
   { actps = tree->GetExpr(index);
     st = actps->GetSymbolType();
     switch(st)       
     {
     case Terminal :                              
     case Empty    : break;
     case NonTerm  :
	  actexpr = actps->GetParseExpr();
	  nt = actexpr->GetNonTermType();
	  switch(nt)
	  {

	   case tformalPar_ssl : isglobalParDef = 1;        // Abschnitt PARAMETERS                 
                                 dummy = AppendType_formalPar_ssl(actexpr);
                                 isglobalParDef = 0; 
                                break;
	   case tIncludeParts :  Handle_IncludeParts(actexpr);   // Abschnitt INCLUDE
          			 break;
	   case tConstantDefinition : Handle_ConstantDefinitions(actexpr, tConstantDefinition);
				      break; // Abschnitt CONSTANTS
	   case tBooleanDefinition :  Handle_BoolDefinitions(actexpr, tBooleanDefinition );
			              break;    // Abschnitt BOOLEANS
	   case tPredicateDefinition : Handle_BoolDefinitions(actexpr, tPredicateDefinition );
					break;  // Abschnitt PREDICATES
	   case tStateFunctionDefinition : Handle_RestDefinitions(actexpr,tStateFunctionDefinition);
					break;  // Abschnitt STATE FUNCTIONS
           case tActionDefinition :  Handle_BoolDefinitions(actexpr,  tActionDefinition );
                                        break;  //Abschnitt ACTIONS
           case tTransitionFunctionDefinition : Handle_RestDefinitions(actexpr, tTransitionFunctionDefinition );
                                        break;  // Abschnitt TRANSITION FUNCTION
           case tAssumptionDefinition : Handle_BoolDefinitions(actexpr,  tAssumptionDefinition );
                                        break;  // Abschnitt ASSUMPTIONS
           case tTheoremDefinition : Handle_BoolDefinitions(actexpr,  tTheoremDefinition ); 
                                        break;  // Abschnitt THEOREMS
           case tTemporalDefinition : Handle_BoolDefinitions(actexpr,  tTemporalDefinition );
                                        break;   // Abschnitt TEMPORALS
           
           default : TypeChecker(actexpr);
                     break;

           }; // end of switch2
           break;
         }; // end of switch1  
        }; // end of for
      
  }



// Behandlung der Produktionsregel 'IncludeParts'

void Handle_IncludeParts(ParseExpr* expr)
 {  assert(expr);
  
    ParseExpr *pe = expr->GetPE(1);
    SymList *symli;
    noTerminal_Type nt = pe->GetNonTermType();
    if(nt == tIncludeParts)
     {  Handle_IncludeParts(pe);
        symli = Handle_IncludePart(expr->GetPE(2));
        if(symli)
         Handle_SubstitutionPart01(expr->GetPE(3),symli);
        isModuleParameter = 0;
     }
    else
     {
       symli = Handle_IncludePart(expr->GetPE(1));
       if(symli)
        Handle_SubstitutionPart01(expr->GetPE(2),symli);
       isModuleParameter = 0;
     };
 }

// Behandlung der Produktionsregel 'IncludePart'

SymList*  Handle_IncludePart(ParseExpr* expr)
 {  assert(expr);
  
    SymListElem* se;
    SymList* symli;
    TypeComp *typ1, *typ2;
    if(expr->GetMaxIndex() == 3)
     { if(expr->GetPE(4)->GetMaxIndex() == 2)
        { se = expr->GetPE(3)->GetSE(1);
           

          isModuleParameter = 1;     // Definition von Modulparametern
          
         
          typ1 = AppendType_formalPar_ssl(expr->GetPE(4)->GetPE(2),1);
    
          typ2 = new TypeComp(OPERATOR_type);
          typ2->SetComp1(typ1);
          typ2->SetComp2(new TypeComp(ANY_type));
          AddTypeInfo(se, typ2); 
        };
     };
    symli = expr->GetIncludes();
    return symli;
  };

// Behandlung der Produktionsregel 'SubstitutionPart01'

void Handle_SubstitutionPart01(ParseExpr* expr,SymList* sl)
 { assert(expr && sl);
  
   if(expr->GetMaxIndex() == 1) // -> Regel nicht leer !!
    {
       Handle_Substitution(expr->GetPE(2), sl);
    };
 }


// Behandlung der Produktionsregel 'Substitution'

void Handle_Substitution(ParseExpr* expr,SymList* l)
 { assert(expr && l);
  
   SymListElem *se, *se_err;
   TypeComp *typ1,*typ2;
   ParseExpr *pe = expr->GetPE(1);
   noTerminal_Type nt = pe->GetNonTermType();
   if(nt == tOneSubstitution_)
    { typ1 = CreateType(pe->GetPE(3));
      se_err = pe->GetSE(1);
      se = pe->GetSE(1)->GetTypeDefinition();
      if(se != NULL)
       {
           typ2 = se->GetTypeInfo();
         
           if(typ2 != NULL) 
            { 
            
              if(typ2->GetType() == UNKNOWN_type)
               { 
                 SimpleWarning("Warning: ID has unknown type !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line); 
               
               }
              else if(typ2->GetType() != ANY_type)
               {

                 SimpleWarning("Warning: ID has a concrete type !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);
                 SimpleWarningCont(" --->");
                 typ2->PrintType_nl();
                } 
                               
               ReplacePar(pe->GetSE(1),GetElementofType(typ1), l);
             
              }
            else
              { 
                SimpleWarning("ID has no type information !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);
              }
          }
     else
        { 
           SimpleWarning("ID has no type definition !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);
        }
 
      }
   else
    { Handle_Substitution(pe, l);
      typ1 = CreateType(expr->GetPE(3)->GetPE(3));
      se_err = expr->GetPE(3)->GetSE(1);
      se = expr->GetPE(3)->GetSE(1)->GetTypeDefinition();
      if(se != NULL)
       {
         typ2 = se->GetTypeInfo();
         if(typ2 != NULL)
          {
            if(typ2->GetType() == UNKNOWN_type)
             {
                SimpleWarning("Warning: ID has unknown type !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);

             }
            else if(typ2->GetType() != ANY_type)
              {

               SimpleWarning("Warning: ID has a concrete type !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);
               SimpleWarningCont(" --->");
               typ2->PrintType_nl();
              }
             

            ReplacePar(expr->GetPE(3)->GetSE(1), GetElementofType(typ1), l);
           }
        else
         {
          SimpleWarning("ID has no type information !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);
          }
       }
    
     else
        {
           SimpleWarning("ID has no type definition !",
                                      se_err->pos,
                                      se_err->leng,
                                      se_err->line);
        }

    }
 }



void ReplacePar(SymListElem* par, TypeComp* typ,SymList* inclist)
 { assert(par && typ && inclist);
  
   // Parametersubstitution innerhalb der Typinformationen der Bezeichner
   inclist->SubstPar_with_Type(par, typ);
    
 }


// Behandlung der Produktionsregel 'expr_list'

void Handle_expr_list(ParseExpr* expr)
{ assert(expr);
  
  ParseExpr *pe = expr->GetPE(1);
  SymListElem *se, *se_err;
  int ok;
  TypeComp *typ1, *typ2, *typ3;
  noTerminal_Type nt = pe->GetNonTermType();
  if(nt == tSymbolDef_LHS)
   { se = AppendType_SymbolDef_LHS(expr->GetPE(1), texpr_list);
     typ1 = CreateType(expr->GetPE(3));
     se_err = expr->GetSE(2);
     if(expr->GetPE(1)->GetMaxIndex() == 1)
     {
       if(se->GetTypeInfo()->GetType() == JOKER_type)
       {  delete se->GetTypeInfo(); // loesche alten Typ des Bezeichners
          AddTypeInfo(se, typ1); // und fuege den Typ der Definition an
       }
       else
       { 
         typ2 = se->GetTypeInfo();
         
         ok = CompareTypes(typ1, typ2,1);
         if(!ok)
          {
            SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
            SimpleWarningCont(" type on the right side:\n ");
            typ1->PrintType_nl();
            
            SimpleWarningCont(" type on the left side:\n ");
            typ2->PrintType_nl();
            
          };  
         delete typ1;typ1 = NULL;
         
         
       };
     }
     else
     { typ2 = se->GetTypeInfo()->GetComp2();
       if(typ2->GetType()==JOKER_type)
         typ2->SetType(ANY_type);
         
       typ3 = typ2->Copy();;
       ok = CompareTypes(typ1, typ3,1);
       if(!ok)
        {
          SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
          SimpleWarningCont(" type on the right side:\n ");
          typ1->PrintType_nl();
          SimpleWarningCont(" type on the left side:\n ");
          typ3->PrintType_nl();
         
        }
       delete typ1;typ1 = NULL;
       delete typ3;typ3 = NULL;
          
     };
    }
   else
    { 
      se_err = expr->GetSE(3);
      Handle_expr_list(pe);
      se = AppendType_SymbolDef_LHS(expr->GetPE(2), texpr_list);
     typ1 = CreateType(expr->GetPE(4));

     if(expr->GetPE(2)->GetMaxIndex() == 1)
     {
         if(se->GetTypeInfo()->GetType() == JOKER_type)
       
       {  delete se->GetTypeInfo();  // loesche alten Typ des Bezeichners
          AddTypeInfo(se, typ1); // und fuege den Typ der Definition an
       }
       else
       { 
         typ2 = se->GetTypeInfo();
         
         ok = CompareTypes(typ1, typ2,1);
         if(!ok)
          { 
           SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
           SimpleWarningCont(" type on the right side:\n ");
           typ1->PrintType_nl();
           SimpleWarningCont(" type on the left side:\n ");
           typ2->PrintType_nl();
           
          }
         delete typ1; typ1 = NULL;
       };
     }
     else
     { typ2 = se->GetTypeInfo()->GetComp2();
       if(typ2->GetType() == JOKER_type)
         typ2->SetType(ANY_type);
       typ3 = typ2->Copy(); 
       ok = CompareTypes(typ1, typ3,1);
       if(!ok)
        {
           SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
           SimpleWarningCont(" type on the right side:\n ");
           typ1->PrintType_nl();
           SimpleWarningCont(" type on the left side:\n ");
           typ3->PrintType_nl();
           
        }
       delete typ1; typ1 = NULL; 
       delete typ3; typ3 = NULL;
      
     };
   };
}

// Behandlung der Produktionsregel 'ConstantDefinitions'

void Handle_ConstantDefinitions(ParseExpr* expr,noTerminal_Type nt)
 { assert(expr);
  
   TypeComp *typ1, *typ2, *typ3;
   SymListElem *se,*se_err;

   int ok;
   se = AppendType_SymbolDef_LHS(expr->GetPE(1), nt);
   se_err = expr->GetSE(2);
   typ1 = CreateType(expr->GetPE(3));
   

   if(expr->GetPE(1)->GetMaxIndex() == 1)  // wenn kein OPERATOR oder ARRAY mit Parameter
   {  if(se->GetTypeInfo()->GetType() == JOKER_type)  
        
       {
         delete se->GetTypeInfo(); // loesche alten Typ des Bezeichners

         AddTypeInfo(se, typ1); // und fuege den Typ der Definition an
       }
       else
       { 
         
         typ2 = se->GetTypeInfo();
         ok = CompareTypes(typ1, typ2,1);
         if(!ok)
          {
            SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
            SimpleWarningCont(" type on the right side:\n ");
            typ1->PrintType_nl();
            SimpleWarningCont(" type on left side:\n ");
            typ2->PrintType_nl();
           
          }
         delete typ1;typ1 = NULL;
       
         
       };
    }
   else  // OPERATOR oder ARRAY mit Parameter
    {
      
      typ2 = se->GetTypeInfo()->GetComp2();
      if(typ2->GetType() == JOKER_type)
        typ2->SetType(ANY_type);
       typ3 = typ2->Copy();
       ok = CompareTypes(typ1,typ3,1);
       if(!ok)
        {
           SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
           SimpleWarningCont(" type on the right side:\n ");
           typ1->PrintType_nl();
           SimpleWarningCont(" type on the left side:\n ");
           typ3->PrintType_nl();
           
        }
       delete typ1; typ1 = NULL;
       delete typ3; typ3 = NULL;
      
    };

 }


void Handle_BoolDefinitions(ParseExpr* expr,noTerminal_Type nt)
 { assert(expr);
  
   TypeComp *typ1, *typ2, *typ3;
   SymListElem *se, *se_err;
   se = AppendType_SymbolDef_LHS(expr->GetPE(1),nt);
   typ1 = CreateType(expr->GetPE(3));
   
   int ok;
   
   se_err = expr->GetSE(2);
   if(expr->GetPE(1)->GetMaxIndex() == 1)  // wenn kein OPERATOR oder ARRAY mit Parameter
   {  typ2 = se->GetTypeInfo();
      
      ok = CompareTypes(typ1, typ2,1);
      if(!ok)
       { 
          SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
          SimpleWarningCont(" type on the right side:\n ");
          typ1->PrintType_nl();
          SimpleWarningCont(" type on the left side:\n ");
          typ2->PrintType_nl();
         
       }
      delete typ1; typ1 = NULL;
    }
   else  // OPERATOR oder ARRAY mit Parameter
    { typ2 = se->GetTypeInfo()->GetComp2();
      typ3 = typ2->Copy();
      ok = CompareTypes(typ1,typ3,1);
      if(!ok)
       {
          SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
          SimpleWarningCont(" type on the right side:\n ");
          typ1->PrintType_nl();
          SimpleWarningCont(" type on the left side:\n ");
          typ3->PrintType_nl();
        

       }
      delete typ1; typ1 = NULL;
      delete typ3; typ3 = NULL;
    };
 }

void Handle_RestDefinitions(ParseExpr* expr,noTerminal_Type nt)
 { assert(expr);
  
   TypeComp *typ1, *typ2,*typ3 ;
   SymListElem *se, *se_err;
   se = AppendType_SymbolDef_LHS(expr->GetPE(1),nt);
   typ1 = CreateType(expr->GetPE(3));
   
   int ok;
   
   se_err = expr->GetSE(2);
   if(expr->GetPE(1)->GetMaxIndex() == 1)  // wenn kein OPERATOR oder ARRAY mit Parameter
   {  typ2 = se->GetTypeInfo();
      if(typ2->GetType() == JOKER_type)
       typ2->SetType(ANY_type);
      ok = CompareTypes(typ1, typ2,1);
      if(!ok)
       {
          SimpleWarning("incompatible types", se_err->pos, se_err->leng, se_err->line);
          SimpleWarningCont(" type on the right side:\n ");
          typ1->PrintType_nl();
          SimpleWarningCont(" type on the left side:\n ");
          typ2->PrintType_nl();  
          
       }
      delete typ1; typ1 = NULL;
      
    }
   else  // OPERATOR oder ARRAY mit Parameter
    { typ2 = se->GetTypeInfo()->GetComp2();
      if(typ2->GetType() == JOKER_type)
       typ2->SetType(ANY_type);
      typ3 = typ2->Copy();
      ok = CompareTypes(typ1,typ3,1);
      if(!ok)
       {
          SimpleWarning("Warning: incompatible types", se_err->pos, se_err->leng, se_err->line);
          SimpleWarningCont(" type on the right side:\n ");
          typ1->PrintType_nl();
          SimpleWarningCont(" type on the left side:\n ");
          typ3->PrintType_nl();
          
       }
      delete typ1; typ1 = NULL;
      delete typ3; typ3 = NULL;
    };
 }

// Behandlung der Produktionsregel 'SymbolDef_LHS'

SymListElem* AppendType_SymbolDef_LHS(ParseExpr* expr, noTerminal_Type nt )
 {   assert(expr);
  
     TypeComp *typ1, *typ2, *typ3;
     ParseExpr *pe1;
     SymListElem *se1,*se2 = 0 ;
     ParseSymb *ps = expr->GetExpr(2);
     Symbol_Type st = ps->GetSymbolType();

     switch(st)
     {
       case NonTerm : se1 = expr->GetSE(1);
              
		      typ1 = CreateType(expr->GetPE(2));
              
		      if(((nt == tBooleanDefinition) || (nt == tPredicateDefinition) ||
			 (nt == tActionDefinition) || (nt == tAssumptionDefinition) ||
			 (nt == tTheoremDefinition) || (nt == tTemporalDefinition)) &&
                         (typ1->GetType() == JOKER_type))
                        {
                          typ1->SetType(BOOLEAN_type);
                           AddTypeInfo(se1,typ1);
                        }
                  
                      else if(typ1->GetType() != JOKER_type)

		        {
                         AddTypeInfo(se1, GetElementofType(typ1));
                         delete typ1; typ1 = NULL;
                        }
                      else
                       AddTypeInfo(se1,typ1);
                  
		  return se1;
       case Terminal : se1 = expr->GetSE(2);
		       if(se1->tokenID == RBR_ON) // Operator-Typ
               { typ1 = new TypeComp(OPERATOR_type);
                 typ2 = AppendType_formalPar_ssl(expr->GetPE(3), 1);
                 
                 typ3 = CreateType(expr->GetPE(5));
              
                 typ1->SetComp1(typ2);
                 typ1->SetComp2(typ3);
                 se2 = expr->GetSE(1);
                 AddTypeInfo(se2, typ1);
                }
                else if(se1->tokenID == '[')  // ARRAY-Typ
                { typ1 = new TypeComp(ARRAY_type);
                  typ2 = AppendType_formalPar_ssl(expr->GetPE(3), 1);
                 
                 
                  typ3 = CreateType(expr->GetPE(5));
              
                  typ1->SetComp1(typ2);
                  typ1->SetComp2(typ3);
                  se2 = expr->GetSE(1);
                  AddTypeInfo(se2, typ1);
                }
                else  // infix OPERATOR-Typ
                { typ1 = new TypeComp(OPERATOR_type);
                  typ2 = AppendType_formalPar(expr->GetPE(7),1);
                  typ2 = AppendType_formalPar(expr->GetPE(3),1,typ2);
                 
                  typ3 = CreateType(expr->GetPE(9));
                  typ1->SetComp1(typ2);
                  typ1->SetComp2(typ3);
                  se2 = expr->GetSE(5);
                  AddTypeInfo(se2, typ1);
                };
               if(((nt == tBooleanDefinition) || (nt == tPredicateDefinition) ||
                   (nt == tActionDefinition) || (nt == tAssumptionDefinition) ||
                   (nt == tTheoremDefinition) || (nt == tTemporalDefinition)) &&
                   (typ3->GetType() == JOKER_type))
                 { 
                   typ3->SetType(BOOLEAN_type);
                  
                 }
                else if(typ3->GetType() != JOKER_type)
                 {
                   typ3 = GetElementofType(typ3);
                   delete typ1->GetComp2();
                   typ1->SetComp2(typ3);
                 } 
 
               return se2;



       }
   }               


int GetSelector(SymListElem* selector)
{ assert(selector);
  
  long num;
  char *string = new (char[strlen(selector->Getlexem())+1]);
  strcpy(string, selector->Getlexem());
  string[0] = ' '; // Punkt aus dem String nehmen!!
  
  num = strtol(string,NULL,10);
  delete string;
  return (int) num;
}



// Behandlung der Produktionsregel 'expr_csl'

TypeComp* GetType_expr_csl(ParseExpr* expr,TypeComp  *prev = NULL) // wird auch fuer constExpr_csl
 {  assert(expr);                                               // genutzt!
  
    ParseExpr *pe = expr->GetPE(1);
    TypeComp* type;
    noTerminal_Type nt = pe->GetNonTermType();
    if((nt == tconstExpr_csl) || (nt == texpr_csl))
      {  type = CreateType(expr->GetPE(3));
	 type->SetComp3(prev);
	 return GetType_expr_csl(pe,type);
      }
    else
      {
	type = CreateType(pe);
	type->SetComp3(prev);
	return type;
      };
 }



 
// Behandlung der Produktionsregel 'qualifierPrefixes'

TypeComp*  CheckType_qualifierPrefixes(ParseExpr* expr,TypeComp* idtyp)
 {   assert(expr && idtyp);
  
     int ok , tokid;
     char* parname = NULL;
     TypeComp *typ1, *typ2,*typ3,*typ4,*dummy;
     SymListElem *se,*se_err,*tmpse;
     ParseSymb *ps = expr->GetExpr(1);
     Symbol_Type st = ps->GetSymbolType();
     if(st == NonTerm)
      {
       tokid = expr->GetSE(3)->tokenID;
       if(tokid == RBR_ON)
        {
         se = expr->GetSE(2)->GetTypeDefinition();
         
         se_err = expr->GetSE(2);
         if(se && se->GetTypeInfo() && (se->GetTypeInfo()->GetType()== OPERATOR_type))
          { typ1 = se->GetTypeInfo()->GetComp1();
            typ2 = GetType_expr_csl(expr->GetPE(4));
            typ3 = typ1; typ4 = typ2;
            ok = CompareTypeChain(typ2,typ1,1);
            if(!ok)
             {
                SimpleWarning("call of (module-)OPERATOR_type with incompatible parameter type",
                               expr->GetSE(3)->pos,
                              (int) (expr->GetSE(5)->pos - expr->GetSE(3)->pos + 1),
                              se_err->line);
                SimpleWarningCont(" used parameter type(s):\n ");
                typ2->PrintParameters();
                SimpleWarningCont("\n defined parameter type(s):\n ");
                typ1->PrintParameters();
                SimpleWarningCont("\n");
             }
            while(typ3 && typ4)
            {             
               parname = typ3->GetParname();
               if(typ4->GetType() != JOKER_type)
                {
                  dummy = GetElementofType(typ4);
                  idtyp->TraverseTypeInfo(parname, dummy);
                  delete dummy;
                }
               else
                idtyp->TraverseTypeInfo(parname, typ4);
               
               typ3 = typ3->GetNextType();
               typ4 = typ4->GetNextType();
             };
          }
         else
          {
          
          SimpleWarning("qualifier is not of type (module-)OPERATOR",
                       se_err->pos,se_err->leng,se_err->line);
          
           
          };
         };
        return CheckType_qualifierPrefixes(expr->GetPE(1),idtyp);
	  
      }
     else 
       return idtyp;
      
   }

void AddTypeInfo(SymListElem* se, TypeComp* tp)
  {  assert(se && tp);
     
     TypeComp* tmptype;
     tmptype = se->GetTypeInfo();
     if(tmptype)
      {
        if(tmptype->GetType() == ID_type)
          { 
            delete tmptype;
            se->SetTypeInfo(tp);
          }
        else
          se->SetTypeInfo(tp);
      }
     else
       se->SetTypeInfo(tp);  
  }


TypeComp* CreateType(ParseExpr* expr )
  { assert(expr);
    
    ParseExpr* pe;
    ParseSymb *ps1, *ps2;
    TypeComp* type;
    noTerminal_Type notermtype = expr->GetNonTermType();
    switch(notermtype)
      {
        case tTypeAndSortDecl01 : ps1 = expr->GetExpr(1);
                                  if(ps1->GetSymbolType() == Empty)
                                    { type = new TypeComp(JOKER_type);
                                      return type;
				    }
                                  else
				    { ps1 = expr->GetExpr(2);
                                      pe = ps1->GetParseExpr();
                                      ps2 = pe->GetExpr(1);
                                      pe = ps2->GetParseExpr();
                                      if(pe->GetNonTermType()== tTypeDecl)
                                           { ps1 = pe->GetExpr(1);
                                             type = ComposeType_SetExpr(ps1->GetParseExpr());
					     return type;
					   }
                                      else
                                        { type = new TypeComp(JOKER_type);
                                          return type;
					};
                                     }
        case tSetExpr :  type = ComposeType_SetExpr(expr);
                         return type;

        case texpr :  type = ComposeType_expr(expr);
                      return type;


        case tConstantExpr : ps1 = expr->GetExpr(1);
                             type = ComposeType_expr(ps1->GetParseExpr());
                              return type;

        case tBooleanExpr : ps1 = expr->GetExpr(1);
                             type = ComposeType_expr(ps1->GetParseExpr());
                              return type;

        case tStateFcnExpr : ps1 = expr->GetExpr(1);
                             type = ComposeType_expr(ps1->GetParseExpr());
                              return type;
        case tPredExpr  : ps1 = expr->GetExpr(1);
                             type = ComposeType_expr(ps1->GetParseExpr());
                              return type;
        case tTransFctExpr : ps1 = expr->GetExpr(1);
                             type = ComposeType_expr(ps1->GetParseExpr());
                              return type;
        case tActionExpr : ps1 = expr->GetExpr(1);
                             type = ComposeType_expr(ps1->GetParseExpr());
                              return type;


        default:   assert(0);   
                   
          };
      }

// Behandlung der Produktionsregel 'ID'

TypeComp* GetTypeofID(ParseExpr* expr)
  {  assert(expr);
    
     TypeComp *type,*typ1, *typ2, *dummy;
     char* name;
     ParseExpr* peptr;
     SymListElem *se1,*se2;
     
     se1 = expr->GetSE(2);
     if((se1->Get_nidt() == nidt_rec_comp)&&(expr->GetPE(1)->GetMaxIndex() != 0))
      {  // ID enthaelt Record-Komponentennamen
      
       typ1 = GetTypeofID(expr->GetPE(1));
       name = se1->Getlexem();
        
       type = GetRECORD(typ1, name);
       if(type)
        {
         typ2 = type->Copy();
         
         delete typ1; typ1 = NULL;
         return typ2;
         }
       else
        {
         
          
          SimpleWarning("ID is not of type RECORD or has not this component",
            expr->minpos,
            (int) (expr->maxpos - expr->minpos + 1),
            se1->line);
          SimpleWarningCont(" --->");
          typ1->PrintType_nl();
          
          delete typ1; typ1 = NULL;
          type = new TypeComp(UNKNOWN_type);
          return type;
        };
       }
     else  
       {
         se2 = se1->GetTypeDefinition();
         if(se2)
           { 
              if(se2->GetTypeInfo() != NULL)
               { 
                  type = se2->GetTypeInfo();
                  if(type->GetType() == ID_type)
                    { // Globaler Parametertypname wird erstmalig angewendet
                      peptr = (ParseExpr*) type->GetID_Type();
                      // nachtraegliche Traversierung des Teilbaums 'peptr'
                      dummy = AppendType_formalPar(peptr);

                      type = se2->GetTypeInfo();
                      name = se2->Getlexem();
                      type = type->Copy();
                      type->SetParname(name);                      
                       
                    }
                  else if((se2->Get_nidt() == nidt_formalglobPar) ||
                          ( (se2->Get_nidt() == nidt_formalPar) &&
                           (isModuleParameter) ) 
                         ) 
                    {  // Globaler Parametertypname wurde zuvor schon angewendet
                       // oder Anwendung von formalen (Modul-)Parametern   
                      name = se2->Getlexem();
                      type = se2->GetTypeInfo();

                      type = type->Copy();    
                                          
                      type->SetParname(name);
                    }
                  else
                    { // Anwendung sonstiger Bezeichner ID's

                       type = se2->GetTypeInfo();
                       type = type->Copy();

                    };
           
            }
           else 
         
           {
           type = new TypeComp(UNKNOWN_type);
           SimpleWarning("ID has UNKNOWN type",
                              expr->minpos,
                              (int) (expr->maxpos - expr->minpos + 1),
                              se1->line);
         };
       }
      else
       { 
         type = new TypeComp(UNKNOWN_type);
         SimpleWarning("ID has no type definition",
                          expr->minpos,          
                          (int) (expr->maxpos - expr->minpos +1),
                            se1->line);
       };
    // Evtl. Qualifier vom Bezeichner ueberpuefen
    return CheckType_qualifierPrefixes(expr->GetPE(1),type);
  };   
 }      
        


// Behandlung der Produktionsregel 'rctype_ssl'

TypeComp* AppendRecComp_rctype_ssl(ParseExpr* expr,TypeComp* prev = NULL)
  {  assert(expr);
     
     
     ParseExpr *pe_help, *pe = expr->GetPE(1);
     TypeComp *type;

     if(expr->GetMaxIndex() == 2)
 
       {
           pe_help = expr->GetPE(3);
           type = AppendRecComp_rctype(pe_help,prev);

           return AppendRecComp_rctype_ssl(pe,type);
       }                       
      
     else
       { 
         type = AppendRecComp_rctype(pe,prev);
         
         return type;
       };
  }

// Behandlung der Produktionsregel 'rctype'

TypeComp* AppendRecComp_rctype(ParseExpr* expr, TypeComp* prev = NULL)
  {  assert(expr);
     
     ParseExpr *pe, *pe_help;
     TypeComp *type,dummy;
     pe = expr->GetPE(1);
     pe_help = expr->GetPE(3);
     type = CreateType(pe_help);

     return Connect_RecComps(pe,GetElementofType(type),prev);
  };



TypeComp* Connect_RecComps(ParseExpr* expr,TypeComp* comptype, TypeComp* prev = NULL)
 {   assert(expr && comptype);
     
     
     ParseExpr *pe;
     TypeComp *type;
     SymListElem *se;
     
     if(expr->GetMaxIndex() == 2)
       {
          se = expr->GetSE(3);
          
          type = comptype->Copy();
          type->SetCompname(se->Getlexem());

          type->SetComp3(prev);
          return Connect_RecComps(expr->GetPE(1),comptype,type);
       } 
      else
       {
          se = expr->GetSE(1);
          type = comptype;
          type->SetCompname(se->Getlexem());
          type->SetComp3(prev);
          return type;
      };
 }


// Behandlung der Produktionsregel 'recdef_csl'

TypeComp* AppendRecComp_recdef_csl(ParseExpr* expr,TypeComp* prev = NULL)
  {  assert(expr);
     
     ParseExpr *pe, *pe_help;
     TypeComp *type;
     pe = expr->GetPE(1);
     if(expr->GetMaxIndex() == 2)
      {
         pe_help = expr->GetPE(3);
	 type = AppendRecComp_recdef(pe_help,prev);
	 return AppendRecComp_recdef_csl(pe,type);
      }
     else
      { 
           type = AppendRecComp_recdef(pe,prev);
           return type;
       };
  }


// Behandlung der Produktionsregel 'recdef'

TypeComp* AppendRecComp_recdef(ParseExpr* expr, TypeComp* prev = NULL)
  {  assert(expr);
     
     ParseExpr* pe;
     SymListElem *se = expr->GetPE(1)->GetSE(1);
     TypeComp *type;
     pe = expr->GetPE(3);

     type = CreateType(pe);
     type->SetCompname(se->Getlexem());
     type->SetComp3(prev);
     return type;
  };


// Behandlung der Produktionsregel 'isElementOfSetListQB'

TypeComp* AppendTypes_isElementOfSetListQB(ParseExpr* expr,int isFunc = 0 , TypeComp* prev = NULL)
  {
     assert(expr);
     
     ParseExpr *pe, *pe_help;
     TypeComp *type,*elemtyp, *rettype;
     pe = expr->GetPE(1);
     if(expr->GetMaxIndex() == 2)
       {
	  pe_help = expr->GetPE(3);
          type = CreateType(pe_help);
          elemtyp = GetElementofType(type);
          delete type; type = NULL; 
          
	  rettype  = SetType_normalID_csl(pe ,elemtyp,isFunc,prev);
	  return rettype;
        }
      else
       {
	  pe_help = expr->GetPE(5);
	  type = CreateType(pe_help);
          elemtyp = GetElementofType(type);
          delete type; type = NULL;
	  pe_help  = expr->GetPE(3);
	  type = SetType_normalID_csl(pe_help,elemtyp,isFunc, prev);
	  return AppendTypes_isElementOfSetListQB(pe,isFunc, type);

       };
  }


// Behandlung der Produktionsregel 'normalID_csl'

TypeComp* SetType_normalID_csl(ParseExpr* expr,TypeComp* tp,int isFuncOrOp  = 0,
								   TypeComp* prev = NULL)
{     assert(expr && tp);
     
      
      ParseExpr* pe;
      SymListElem *se;
      TypeComp *rettype,*dummy, *elemtyp, *newtype;
      if(expr->GetMaxIndex() == 2)
       {

	    se = expr->GetSE(3);
            newtype = tp->Copy();
            
            if(isFuncOrOp)  // normalID's sind formale Parameter eines ARRAY- oder OPERATOR-Typs
             {
               if(isModuleParameter)  // Modulparameter
                newtype->SetParname(se->Getlexem());   
            
               AddTypeInfo(se,newtype);
               newtype = newtype->Copy();
               newtype->SetComp3(prev);
              }           
             else
              {
               AddTypeInfo(se, newtype);
               newtype = NULL;
              };
                

             rettype = SetType_normalID_csl(expr->GetPE(1), tp,
                                                            isFuncOrOp, newtype);

	     return rettype;
             
	   
       }
      else
       {
             se = expr->GetSE(1);
             newtype = tp;
            if(isFuncOrOp) // normalID's sind formale Parameter eines ARRAY- oder OPERATOR-Typs
             {
               if(isModuleParameter)   // Modulparameter
                newtype->SetParname(se->Getlexem());

               AddTypeInfo(se,newtype);
               newtype = newtype->Copy();
               newtype->SetComp3(prev);
             }
            else
              {
               AddTypeInfo(se, newtype);
               newtype = NULL;
              };
                    
          return newtype;  
       };
                
 }


// Behandlung der Produktionsregel 'formalPar'

TypeComp* AppendType_formalPar(ParseExpr* expr,int  isFuncOrOp =  0, TypeComp* prev = NULL)
 {  assert(expr);
    
    ParseSymb* ps = expr->GetExpr(1);
    ParseSymb* ps_help;
    ParseExpr *pe, *pe_help;
    TypeComp *partype,*type,*nidtype,*rettype,*typ1,*elemtyp;
    SymListElem* se;
    noTerminal_Type notermtype;
    Symbol_Type symtype = ps->GetSymbolType();
    switch(symtype)
      {
	 case NonTerm : pe = expr->GetPE(1);
                        notermtype = pe->GetNonTermType();
                        if(notermtype == tnormalID_csl)
                          {
                            if(isglobalParDef)  // globale Parameter
                              {                 // -> Teilbaum spaeter untersuchen
                                type = new TypeComp(ID_type); 
                                type->SetID_Type(expr);
                              }
                            else
                              {
                                pe_help = expr->GetPE(2);
                                type = CreateType(pe_help);
                                if(type->GetType() == JOKER_type)
                                 type->SetType(ANY_type);
                                else
                                 type = GetElementofType(type);
                              }; 
                            
                            rettype = SetType_normalID_csl(pe,type, isFuncOrOp , prev);
                            return rettype;
			  }
                        else {
                              se = expr->GetSE(5);
                              if(isglobalParDef)  // globale Parameter
                                                  // -> Teilbaum spaeter untersuchen
                               { nidtype = new TypeComp(ID_type);
                                 nidtype->SetID_Type(expr);
                                 AddTypeInfo(se, nidtype);
                                 return NULL;
                                 
                               }
                              else
                               {
                                nidtype = new TypeComp(OPERATOR_type);  // infix Operator
                                pe_help = expr->GetPE(7);
                                partype = AppendType_formalPar(pe_help,1);
                                pe_help = expr->GetPE(3);
                                partype = AppendType_formalPar(ps_help->GetParseExpr(),1,partype);
                                nidtype->SetComp1(partype);
                              

                                pe_help = expr->GetPE(9);
                              

                               
                                type = CreateType(pe_help);
                                if(type->GetType() == JOKER_type)
                                 type->SetType(ANY_type);
                                                             
			        nidtype->SetComp2(GetElementofType(type));
                                AddTypeInfo(se, nidtype);
                                if(isFuncOrOp)
                                  {
                                   rettype = GetElementofType(type);
                                   rettype->SetComp3(prev);
                                  }
                                else
                                  rettype = NULL;
                                return rettype; 
                              }
                       
	                    }

         case Terminal :   se = expr->GetSE(1);
                           if(isglobalParDef)  // globaler Parameter ...
                            {
                              nidtype = new TypeComp(ID_type);
                              nidtype->SetID_Type(expr);
                              AddTypeInfo(se,nidtype);
                              return NULL;
                             
                                                               
                            }
                           else
                            { 
                             nidtype = new TypeComp(OPERATOR_type);   
                             pe_help = expr->GetPE(3);
                             partype = AppendType_formalPar_ssl(ps_help->GetParseExpr(),1);
                        
                             nidtype->SetComp1(partype);
			  
                             pe_help = expr->GetPE(5);
 
                             type = CreateType(pe_help);
                             if(type->GetType() == JOKER_type)
                              type->SetType(ANY_type);
                        
                             nidtype->SetComp2(GetElementofType(type));
                            
                             AddTypeInfo(se,nidtype);
                           
                             if(isFuncOrOp)
                              {
                                rettype = GetElementofType(type);
                                rettype->SetComp3(prev);
                              } 
                             else
                              rettype = NULL;
                           
                             return rettype;
			    } 

      };
 }
          

// Behandlung der Produktionsregel 'formalPar_ssl'

TypeComp* AppendType_formalPar_ssl(ParseExpr* expr, int isFuncOrOp =0,TypeComp* prev = NULL)
 {   assert(expr);
    
     ParseExpr *pe = expr->GetPE(1), *pe_help;
     TypeComp *type1,*type2;
     noTerminal_Type notermtype;
     
     notermtype = pe->GetNonTermType();
     switch(notermtype)
            {
              case tformalPar_ssl : pe_help = expr->GetPE(3);
                                    type1 = AppendType_formalPar(expr->GetPE(3), isFuncOrOp,prev);
                                    type2 = AppendType_formalPar_ssl(pe, isFuncOrOp,type1);
                                    return type2;

              case tformalPar     : type1 = AppendType_formalPar(pe, isFuncOrOp, prev);
                                    return type1;


	    };
 }
      





// Behandlung der Produktionsregel 'prefixLines'

TypeComp* GetType_prefixLines(ParseExpr* expr)
{ assert(expr);
  
  int ok;
  TypeComp* type;
  ParseExpr *pe = expr->GetPE(1);
  type = GetType_con_dis_Lines(pe);
  
  
  return type;
}

// Behandlung der Produktionsregel 'con_dis_Lines'

TypeComp* GetType_con_dis_Lines(ParseExpr* expr)
{ assert(expr);
  
  int ok;
  SymListElem *se_err;
  TypeComp *type, *typ1, *dummy;
  ParseSymb* ps = expr->GetExpr(1);
  Symbol_Type st = ps->GetSymbolType();
  switch(st)
  {
   case Terminal : se_err = expr->GetSE(1);
                   type = new TypeComp(BOOLEAN_type);
                   typ1 = CreateType(expr->GetPE(2));
                   ok = CompareTypes(typ1, type);
                   if(!ok)
                    {
                      SimpleWarning("expr is not of type BOOLEAN",
                        expr->GetPE(2)->minpos,
                        (int) (expr->GetPE(2)->maxpos - expr->GetPE(2)->minpos + 1),
                        se_err->line);
                      SimpleWarningCont(" --->");
                      typ1->PrintType_nl();
                      
                    } 
                   delete typ1; typ1 = NULL;
                   return type;           
   case NonTerm :  
                   se_err = expr->GetSE(2);
                   
                   type = new TypeComp(BOOLEAN_type);
                   dummy = GetType_con_dis_Lines(expr->GetPE(1));
                   typ1 = CreateType(expr->GetPE(3));
                   ok = CompareTypes(typ1,type);
                   if(!ok)
                    {
                      SimpleWarning("expr is not of type BOOLEAN",
                      expr->GetPE(3)->minpos,
                      (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos + 1),
                       se_err->line);
                      SimpleWarningCont(" --->");
                      typ1->PrintType_nl();
                      
                    }
                   delete dummy; dummy = NULL;
                   delete typ1; typ1 = NULL;
                   return type ;
   default    : assert(0);
  }
}

// Behandlung der Produktionregel 'CaseRA_List'

TypeComp* GetType_CaseRA_List(ParseExpr* expr)
{ assert(expr);
  
  int ok;
  TypeComp *typ1,*typ2;
  ParseExpr *pe = expr->GetPE(1);
  if(pe->GetNonTermType() == tCaseRA)
  { typ1 = GetType_CaseRA(pe);
    return typ1;   
  }
  else
  { typ1 = GetType_CaseRA_List(pe);
    typ2 = GetType_CaseRA(expr->GetPE(2));
    // Typen koennen verschieden sein
    ok = CompareTypes(typ1, typ2);
     
    delete typ2;  typ2 = NULL;
    return typ1;
  }
}

// Behandlung der Produktionsregel 'CaseRA'

TypeComp* GetType_CaseRA(ParseExpr* expr)
{ assert(expr);
  
  int ok;
  SymListElem *se_err = expr->GetSE(2);
  TypeComp *type = new TypeComp(BOOLEAN_type), *typ1, *typ2;
  typ1 = CreateType(expr->GetPE(1));
  ok = CompareTypes(typ1, type);
  if(!ok)
   {
     SimpleWarning("expr is not of type BOOLEAN",
                    expr->GetPE(1)->minpos,
                    (int) (expr->GetPE(1)->maxpos - expr->GetPE(1)->minpos + 1),
                    se_err->line);
     SimpleWarningCont(" --->");
     typ1->PrintType_nl(); 
     
   }
  delete type; type = NULL;
  delete typ1; typ1 = NULL;
  typ2 = CreateType(expr->GetPE(3));
  return typ2;
}
    

// Behandlung der Produktionsregel 'SetExpr'

TypeComp* ComposeType_SetExpr(ParseExpr* expr)
{ assert(expr);
  
  ParseSymb *ps = expr->GetExpr(1), *ps_help;
  Symbol_Type st = ps->GetSymbolType();
  ParseExpr *pe, *pe_help; 
  noTerminal_Type nt;
  TypeComp *type , *typ1 ,*typ2, *typ3, *dummy;
  SymListElem *se, *se_err;
  int ok,number;
  switch(st)
  { case NonTerm : pe = expr->GetPE(1);
		   nt = pe->GetNonTermType();
		   switch(nt)
		   {case tID : if(expr->GetMaxIndex()== 0) // ID
			       { type = GetTypeofID(pe);
				 return type;
                               }
			
			       else                      // ID [ ]
			       { se_err = expr->GetSE(2);
                                  typ2 = GetType_expr_csl(expr->GetPE(3));
                                 type = GetTypeofID(pe);
				 typ3 = GetARRAY(type);
				 if(typ3 != NULL)
				 { typ1 = type->GetComp1();
				   
				   ok = CompareTypeChain(typ2, typ1,1);
                                   if(!ok)
                                    { 
                                    SimpleWarning("application of type ARRAY with incompatible parameter list",
                                                     expr->GetSE(2)->pos ,
                                                     (int) (expr->GetSE(4)->pos - expr->GetSE(2)->pos+ 1),
                                                     se_err->line);
                                    SimpleWarningCont(" used parameter type(s):\n ");
                                    typ2->PrintParameters();
                                    SimpleWarningCont("\n defined parameter type(s):\n ");
                                    typ1->PrintParameters();
                                    SimpleWarningCont("\n");
                                    }
                                   delete typ2; typ2 = NULL;
				   typ3 = typ3->GetComp2();
                                   typ2 = typ3->Copy();
                                   delete type; type = NULL;
				   return typ2;
				 }
                                         
				 else // Kein ARRAY!
				 { 
                                   delete typ2; typ2 = NULL;
                                   
                                   SimpleWarning("ID is not of type ARRAY",
                                                    expr->GetPE(3)->minpos,
                                                    (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos+ 1),
                                                    se_err->line);
                                   SimpleWarningCont(" --->");
                                   type->PrintType_nl();
                                  
                                   delete type; type = NULL;
                                   type = new TypeComp(UNKNOWN_type);
				   return type;
				 }
				
			       }
		    case tOperatorApplication :      // OperatorApplication
			     type = GetTypeofID(pe);
                             
                              typ2 = GetType_expr_csl(pe->GetPE(4));
                            if(type->GetType() == OPERATOR_type)
			     { typ1 = type->GetComp1();
			       se_err = expr->GetPE(1)->GetSE(3);
			       ok = CompareTypeChain(typ2, typ1,1);
                               if(!ok)
                                {
                                SimpleWarning("application of type OPERATOR with incompatible parameter list",
                                                 pe->GetSE(3)->pos ,
                                                 (int) (pe->GetSE(5)->pos - pe->GetSE(3)->pos +1),
                                                 se_err->line);
                                SimpleWarningCont(" used parameter type(s):\n ");
                                typ2->PrintParameters();
                                SimpleWarningCont("\n defined parameter type(s):\n ");
                                typ1->PrintParameters();
                                SimpleWarningCont("\n");   

                                }  
                               delete typ2; typ2 = NULL;
			       typ1 = type->GetComp2();
                               typ2 = typ1->Copy();
                               delete type; type = NULL;
			       return typ2;
			      }
        
                            else  // Kein OPERATOR!
			      { 
                                delete typ2; typ2 = NULL;
                                se_err = expr->GetPE(1)->GetSE(2);
                                
                                SimpleWarning("ID not of type OPERATOR",
                                                 se_err->pos, se_err->leng, se_err->line);
                                SimpleWarningCont(" --->");
                                type->PrintType_nl();
                               
                                 
                                delete type; type = NULL;
                                type = new TypeComp(UNKNOWN_type);
				return type;
			      }
                            
		    case tSetExpr : se = expr->GetPE(2)->GetSE(1);    // SetExpr setOp SetExpr
                                   typ1 = CreateType(pe);
                                   typ2 = CreateType(expr->GetPE(3));
			     if(se->tokenID == _UNION_)   
			     { if((typ1->GetType() == UNION_type) && (typ2->GetType() == UNION_type))
                               { typ3 = GetLastType(typ1->GetComp1());
                                 typ3->SetComp3(typ2->GetComp1());
                                 return typ1;
			       }
                               else if(typ1->GetType() == UNION_type)
                               { typ3 = GetLastType(typ1->GetComp1());
                                 typ3->SetComp3(typ2);
                                 return typ1;
                               }
                               else if(typ2->GetType() == UNION_type)
                               {
                                 type = new TypeComp(UNION_type);
                                 type->SetComp1(typ1);
                                 typ1->SetComp3(typ2->GetComp1());
                                 return type;
                               }
                               else
                               { type = new TypeComp(UNION_type); 
			         typ1->SetComp3(typ2);
			         type->SetComp1(typ1);
			         return type;
                               };
			     }
			     else if(se->tokenID == _CROSS_)
			     { if((GetTUPLE(typ1->GetComp1()) != NULL) && (GetTUPLE(typ2->GetComp1()) != NULL))
                                { typ3 = GetLastType(typ1->GetComp1()->GetComp1());
                                  typ3->SetComp3(typ2->GetComp1()->GetComp1());
                                  return typ1;
                                }
                               else if(GetTUPLE(typ1->GetComp1())!= NULL)
                                { typ3 = GetLastType(typ1->GetComp1()->GetComp1());
                                  typ2 = GetElementofType(typ2);
                                  typ3->SetComp3(typ2);
                                  return typ1;
                                }
                               else if(GetTUPLE(typ2->GetComp1())!= NULL)
                                { type = new TypeComp(SET_type);
                                  typ3 = new TypeComp(TUPLE_type);
                                  typ1 = GetElementofType(typ1);
                                  typ3->SetComp1(typ1);
                                  typ1->SetComp3(typ2->GetComp1()->GetComp1());
                                  type->SetComp1(typ3);
                                  return type;
                                }
                               else
                                { type = new TypeComp(SET_type);
                                  typ3 = new TypeComp(TUPLE_type);
                                  typ1 = GetElementofType(typ1);
                                  typ2 = GetElementofType(typ2);
                                  typ1->SetComp3(typ2);
			          typ3->SetComp1(typ1);
                                  type->SetComp1(typ3);
			          return type;
                                };
			     }
			     else if((se->tokenID == _INTERSECT_) ||
				     (se->tokenID == _SET_DIFFERENCE_))
			     { se_err = se;
			       
			       ok = CompareTypes(typ2, typ1);
                               if(!ok)
                                 {
                                  SimpleWarning("left and right expr are of incompatible types",
                                                 se_err->pos, se_err->leng, se_err->line);
                                  SimpleWarningCont(" type of the right side:\n ");
                                  typ2->PrintType_nl();
                                  SimpleWarningCont(" type of the left side:\n ");
                                  typ1->PrintType_nl();
                                 
                                 };
                               delete typ2; typ2 = NULL;
			       return typ1;
			     }
		    case tbeginSet : pe_help = expr->GetPE(2);
                             
			     if(pe_help->GetNonTermType() == texpr)  // beginSet expr '|' isElementOfSetListQB
                                                                    
			     { dummy = AppendTypes_isElementOfSetListQB(expr->GetPE(4));
			       typ1 = CreateType(expr->GetPE(2));
			       type = new TypeComp(SET_type);
			       type->SetComp1(typ1);
			       if(expr->GetSE(5)->tokenID == COLON)  // ...COLON SetExpr '}'
			       { se_err = expr->GetSE(5);
                                 typ2 = new TypeComp(BOOLEAN_type);
				 typ1 = CreateType(expr->GetPE(6));
				 ok = CompareTypes(typ1, typ2);
                                 if(!ok)
                                  {
                                  SimpleWarning("right expr is not of type BOOLEAN",
                                                    expr->GetPE(6)->minpos,
                                                    (int) (expr->GetPE(6)->maxpos - expr->GetPE(6)->minpos +1),
                                                    se_err->line);
                                  SimpleWarningCont(" --->");
                                  typ1->PrintType_nl();
                                 
                                  } 
				 delete typ2; typ2 = NULL;
                                 delete typ1; typ1 = NULL;
			       }
                               
			       return type;
			     }
			     else
			     { typ1 = GetType_expr_csl(expr->GetPE(2));
			       type = new TypeComp(ENUMSET_type);
			       type->SetComp1(typ1);
			       return type;
			     }
		    case tpredefSet : se = pe->GetSE(1);   // predefSet
                                      se_err = se;
			     if(se->tokenID == NAT)
			     { type = new TypeComp(SET_type);
                               typ1 = new TypeComp(NATNUM_type);
                               type->SetComp1(typ1);
			       return type;
			     }
			     else // fuer weitere vordef. Typen
			     { type = new TypeComp(UNKNOWN_type);
                               SimpleWarning("unknown predefined type",
                                              se_err->pos,
                                              se_err->leng,
                                              se_err->line);
                               return type;
                             }
		   default : assert(0);
		   }
    case Terminal : se = expr->GetSE(1);

		    switch(se->tokenID)
		    {
		    case ANY : type = new TypeComp(ANY_type);        // ANY
			       return type;
		    case RBR_ON : type = CreateType(expr->GetPE(2)); // RBR_ON SetExpr RBR_OFF
				  return type;
		    case SUBSET : typ1 = CreateType(expr->GetPE(3)); // SUBSET RBR_ON SetExpr RBR_OFF
				  type = new TypeComp(SUBSET_type);
				  type->SetComp1(typ1);
				  return type;
		    case UNION : typ1 = CreateType(expr->GetPE(3));  // UNION RBR_ON SetExpr RBR_OFF
				 type = new TypeComp(SET_type);    // Verknuepfung mehrerer Mengen !
                                 typ1 = GetElementofType(typ1);    // Mengenelemente der Menge
                                 typ1 = GetElementofType(typ1);    // Elemente der Mengenelemente
				 type->SetComp1(typ1);             // Verknuepfung zu einer Menge
				 return type;
		    case _EMPTYSET_ : type = new TypeComp(EMPTYSET_type);  // _EMPTYSET_
				 return type;
		    case SET   : type = new TypeComp(SET_type); // SET '[' SetExpr _RIGHTARROW_ SetExpr ']'      
                                 typ1 = CreateType(expr->GetPE(3));
				 typ2 = CreateType(expr->GetPE(5));
				 typ3 = new TypeComp(ARRAY_type);
				 typ3->SetComp1(GetElementofType(typ1));
				 typ3->SetComp2(GetElementofType(typ2));
                                 type->SetComp1(typ3);
				 return type;


		    case '['   : typ3 = GetTypeofID(expr->GetPE(3)); // '[' CHANGE ID SEMICOLON expr COLON ...
				 type = GetARRAY(typ3->GetComp1());
                                 
                                 if(type)   // falls ARRAY ...
                                 { se_err = expr->GetSE(4);
				   typ1 = type->GetComp1(); // Domain von ARRAY
                                   typ2 = CreateType(expr->GetPE(5)); 
                                  
                                   ok = CompareTypes(typ2, typ1,1);  // stimmt 'expr' Typ mit typ1 ueberein?
                                     
                                   
                                   if(!ok)
                                    {
                                     SimpleWarning("expr is not of ARRAY-domain type",
                                                      expr->GetPE(5)->minpos,
                                                      (int) (expr->GetPE(5)->maxpos - expr->GetPE(5)->minpos +1),
                                                      se_err->line);
                                     SimpleWarningCont(" used domain type:\n ");
                                     typ2->PrintParameters();
                                     SimpleWarningCont(" \ndefined domain type:\n ");
                                     typ1->PrintParameters();
                                     SimpleWarningCont("\n");

                                    }
                                   delete typ2; typ2 = NULL;
                                   typ2 = CreateType(expr->GetPE(7)); // neuer Bildbereichtyp
                                   if(type->GetComp2()->GetType() == UNION_type)
                                    { typ1 = type->GetComp2()->GetComp1();
                                      typ1 = GetLastType(typ1);
                                      typ1->SetComp3(GetElementofType(typ2));
                                    }
                                   else
                                    {
                                    typ1 = new TypeComp(UNION_type); // Bildbereich hat nun 2 Typen!
                                    typ1->SetComp1(type->GetComp2());
                                    typ1->GetComp1()->SetComp3(GetElementofType(typ2));
                                    type->SetComp2(typ1);
                                    };                                    
                                   return typ3;
                                 }
                                 else
                                 { se_err = expr->GetSE(2);  
                                   
                                   SimpleWarning("ID is not of type ARRAY",
                                                    expr->GetPE(3)->minpos,
                                                    (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos +1),
                                                    se_err->line);
                                   SimpleWarningCont(" --->");
                                   typ3->PrintType_nl(); 
                                   
                                   delete typ3; typ3 = NULL;
                                   type = new TypeComp(UNKNOWN_type);
                                   return type;   
                                  
				 }
		    case DOMAINt : se_err = expr->GetSE(2);    // DOMAINt RBR_ON ID RBR_OFF
                                   typ1 = GetTypeofID(expr->GetPE(3));
                                   type = GetARRAY(typ1);
                                   if(type != NULL)
                                    {
                                      type = type->GetComp1();
                                      return type;
                                    }
                                   else
                                    { 
              
                                      SimpleWarning("ID is not of type ARRAY",
                                                       expr->GetPE(3)->minpos,
                                                       (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos +1),
                                                       se_err->line);
                                      SimpleWarningCont(" --->");
                                      typ1->PrintType_nl();
                                    
                                      delete typ1; typ1 = NULL;
                                      type = new TypeComp(UNKNOWN_type);
                                      return type;
                                     }
		   case DSBR_ON : if(expr->GetMaxIndex() == 2)  // DSBR_ON rctype_ssl DSBR_OFF
				  { type = new TypeComp(SET_type);
                                    typ1 = AppendRecComp_rctype_ssl(expr->GetPE(2));
				    typ3 = new TypeComp(RECORD_type);
				    typ3->SetComp1(typ1);
                                    type->SetComp1(typ3);
                                   
				    return type;
				  }
				  else      // DSBR_ON CHANGE ID ...
				  { se_err = expr->GetSE(2);
                                    typ1 = GetTypeofID(expr->GetPE(3)); 
                                    se = expr->GetPE(5)->GetSE(1);
                                    type = GetRECORD(typ2->GetComp1(), se->Getlexem());
                                    if(type != NULL)
				     { typ2 = CreateType(expr->GetPE(7));
                                       typ2 = GetElementofType(typ2);
                                       delete type->GetComp1(); type->SetComp1(NULL);
                                       delete type->GetComp2(); type->SetComp2(NULL);
                                       type->SetType(typ2->GetType());
                                       
                                       if(typ2->GetComp1() != NULL)
                                         type->SetComp1(typ2->GetComp1());
                                       if(typ2->GetComp2() != NULL) 
                                         type->SetComp2(typ2->GetComp2());
                                       if(typ2->GetParname() != NULL)
                                         type->SetParname(typ2->GetParname());
                               
                                       return typ1;
                                     } 
                                    else
                                     {
				       
                                       SimpleWarning("ID is not of type RECORD or has no this compname",
                                                        expr->GetPE(3)->minpos,
                                                        (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos +1),
                                                        se_err->line);
                                       SimpleWarningCont(" --->");
                                       typ1->PrintType_nl();
                                       
                                       delete typ1; typ1 = NULL;
                                       type = new TypeComp(UNKNOWN_type); 
                                       return type;
                                     }
                                   }    
		   default : assert(0);
		    }
    default :  assert(0);
   }
  }


// Behandlung der Produktionsregel 'expr'

TypeComp* ComposeType_expr(ParseExpr* expr)
{ assert(expr);
  
  ParseSymb *ps = expr->GetExpr(1), *ps_help;
  Symbol_Type st = ps->GetSymbolType();
  ParseExpr *pe, *pe_help;
  noTerminal_Type nt;
  int number;
  TypeComp *type , *typ1 ,*typ2, *typ3, *dummy;
  SymListElem *se,*se_err;
  int ok;
  switch(st)
  {
   case Terminal : se = expr->GetSE(1);
	  switch(se->tokenID)
	  {
	  case RBR_ON : pe_help = expr->GetPE(2);   // RBR_ON expr RBR_OFF  
	       if(pe_help->GetNonTermType() == texpr)
	       { type = CreateType(pe_help);
		 return type;
	       }	       
               else       // RBR_ON constExpr_csl RBR_OFF
	       { typ1 = GetType_expr_csl(pe_help);
                 type = new TypeComp(TUPLE_type);
		 type->SetComp1(typ1);
		 return type;
	       }
	  case TRUE :     // TRUE
	  case FALSE : type = new TypeComp(BOOLEAN_type);  // FALSE
		       return type;
	  case _L_NOT_ : se_err = expr->GetSE(1);   // _L_NOT_ IBL expr IBR
	       typ1 = CreateType(expr->GetPE(2));
	       type = new TypeComp(BOOLEAN_type);
	       ok = CompareTypes(typ1, type);
               if(!ok)
                {
                 SimpleWarning("expr is not of type BOOLEAN",
                                 expr->GetPE(2)->minpos,
                                 (int) (expr->GetPE(2)->maxpos - expr->GetPE(2)->minpos + 1),
                                 se_err->line);
                 SimpleWarningCont(" --->");
                 typ1->PrintType_nl();

                }
               delete typ1; typ1 = NULL;
	       return type;
	  case _EXISTS_QUANTOR_ :   // _EXISTS_QUANTOR_ isElementOfSetListQB DOUBLECOLON ...
	  case _ALL_QUANTOR_ :     // _ALL_QUANTOR_ sElementOfSetListQB DOUBLECOLON ...
               se_err = expr->GetSE(3);
	       type = new TypeComp(BOOLEAN_type);
	       dummy = AppendTypes_isElementOfSetListQB(expr->GetPE(2));
	       typ1 = CreateType(expr->GetPE(4));
	       ok = CompareTypes(typ1, type);
               if(!ok)
                {
                 SimpleWarning("expr is not of type BOOLEAN",
                                 expr->GetPE(4)->minpos,
                                 (int) (expr->GetPE(4)->maxpos - expr->GetPE(4)->minpos + 1),
                                 se_err->line);
                 SimpleWarningCont(" --->");
                 typ1->PrintType_nl();
                 
                }
               delete typ1; typ1 = NULL;
               
	       return type;
	  case CHOOSE : se_err = expr->GetSE(6);   // CHOOSE RBR_ON normalID _ELEMENT_OF_ ...
                        type = CreateType(expr->GetPE(5));
                        typ3 = GetElementofType(type);
			AddTypeInfo(expr->GetSE(3), typ3);
			typ1 = new TypeComp(BOOLEAN_type);
			typ2 = CreateType(expr->GetPE(7));
			ok = CompareTypes(typ2, typ1);
                        if(!ok)
                         { 
                          SimpleWarning("right expr not of type BOOLEAN",
                                 expr->GetPE(7)->minpos,
                                 (int) (expr->GetPE(7)->maxpos - expr->GetPE(7)->minpos + 1),
                                 se_err->line);
                          SimpleWarningCont(" --->");
                          typ2->PrintType_nl();
                          
                         } 
                        delete type; type = NULL;
			delete typ1; typ1 = NULL;
                        delete typ2; typ2 = NULL;
			return typ3->Copy();
	  case '[' : if(expr->GetMaxIndex() == 5)  // '[' FUNCTION isElementOfSetListQB ...
		     { typ1 = AppendTypes_isElementOfSetListQB(expr->GetPE(3),1);
		       typ2 = CreateType(expr->GetPE(5));

		       type = new TypeComp(ARRAY_type);
		       type->SetComp1(typ1);
		       type->SetComp2(typ2);
		       return type;
		     }
		     else if(expr->GetMaxIndex() == 7) // '[' CHANGE ID SEMICOLON ...
		     { type = GetTypeofID(expr->GetPE(3));
		       typ3 = GetARRAY(type);
                       if(typ3 != NULL)
			 { 
                           typ1 = CreateType(expr->GetPE(5));
			   typ2 = CreateType(expr->GetPE(7));
                           se_err = expr->GetSE(4);
                          
			   ok =  CompareTypes(typ1,typ3->GetComp1(),1);
                          
                           if(!ok)
                            {
                             SimpleWarning("expr is not of ARRAY domain type",
                                             expr->GetPE(5)->minpos,
                                             (int) (expr->GetPE(5)->maxpos - expr->GetPE(5)->minpos +1),
                                             se_err->line);
                             SimpleWarningCont(" used domain type:\n ");
                             typ1->PrintParameters();
                             SimpleWarningCont(" expected domain type:\n ");
                             typ3->GetComp1()->PrintParameters();
                             SimpleWarningCont("\n");
                            }
                           se_err = expr->GetSE(6);
                          
			   ok = CompareTypes(typ2, typ3->GetComp2(),1) ;
                          
                           if(!ok)
                            {
                             SimpleWarning("expr is not of ARRAY mapping type",
                                             expr->GetPE(7)->minpos,
                                             (int) (expr->GetPE(7)->maxpos - expr->GetPE(7)->minpos + 1),
                                             se_err->line);
                             SimpleWarningCont(" used mapping type:\n ");
                             typ2->PrintParameters();
                             SimpleWarningCont(" expected mapping type (element of):\n ");
                             typ3->GetComp2()->PrintParameters();
                             SimpleWarningCont("\n");
                            }
                           delete typ1; typ1 = NULL;
                           delete typ2; typ2 = NULL;
			   return type;
			 }
			 else
                         {   
                             se_err = expr->GetSE(2);
                             SimpleWarning("ID is not of type ARRAY",
                                               expr->GetPE(3)->minpos,
                                               (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos + 1),
                                               se_err->line);
                             SimpleWarningCont(" --->");
                             type->PrintType_nl();
                            
                             delete type; type = NULL;
			     type = new TypeComp(UNKNOWN_type);
			     return type;
			 }
		       }
		     else if(expr->GetMaxIndex() == 6)  // '[' ActionExpr ']' '_' RBR_ON ...
		     { se_err = expr->GetSE(1);
                       type = new TypeComp(BOOLEAN_type);
		       typ1 = CreateType(expr->GetPE(2));
		       ok = CompareTypes(typ1, type);
                       if(!ok)
                        {
                            SimpleWarning("expr is not of type BOOLEAN",
                                             expr->GetPE(2)->minpos,
                                             (int) (expr->GetPE(2)->maxpos - expr->GetPE(2)->minpos + 1),
                                             se_err->line);
                            SimpleWarningCont(" --->");
                            typ1->PrintType_nl();
                            
                        }    
                       delete typ1; typ1 = NULL;
  		       return type;
		     }
	  case DSBR_ON : if(expr->GetMaxIndex() == 2)   // DSBR_ON recdef_csl DSBR_OFF
			 { typ1 = AppendRecComp_recdef_csl(expr->GetPE(2));
                           type = new TypeComp(RECORD_type);
                           type->SetComp1(typ1);
                           return type;
			 }
			 else
			 { se = expr->GetPE(5)->GetSE(1);  // DSBR_ON CHANGE ID SEMICOLON ...
                           typ1 = GetTypeofID(expr->GetPE(3));
                           type = GetRECORD(typ1, se->Getlexem());
                           if(type != NULL)
                           { se_err = expr->GetSE(6);
                             typ2 = CreateType(expr->GetPE(7));
                          
                             ok = CompareTypes(typ2, type,1) ;
                            
                             if(!ok)
                              {
                               SimpleWarning("RECORD components have different types",
                                               expr->GetPE(7)->minpos,
                                               (int) (expr->GetPE(7)->maxpos - expr->GetPE(7)->minpos + 1),
                                               se_err->line);
                               SimpleWarningCont(" used component type:\n ");
                               typ2->PrintType_nl();
                               SimpleWarningCont(" defined type:\n ");
                               type->PrintType_nl();
                              
                              }
                             delete typ2; typ2 = NULL;
                             return typ1;   
			   }
                           else
                           { se_err = expr->GetSE(2); 
                             SimpleWarning("ID is not of type RECORD",
                                              expr->GetPE(3)->minpos,
                                              (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos + 1),
                                              se_err->line);
                             SimpleWarningCont(" --->");
                             typ1->PrintType_nl();
                           
                             delete typ1; typ1 = NULL;
                             type = new TypeComp(UNKNOWN_type);
                             return type;   
                           }
                         }
//	  case MINUS  : type = new TypeComp(NATNUM_type);    
//			typ1 = CreateType(expr->GetPE(2));
//			ok = CompareTypes(typ1, type);
//
//			return type;
	  case NATNUM : type = new TypeComp(NATNUM_type);  // NATNUM
			return type;
	  case IF     : se_err = expr->GetSE(1);  // IF IBL expr IBR THEN IBL expr ...
                        typ1 = new TypeComp(BOOLEAN_type);
			typ2 = CreateType(expr->GetPE(2));
			ok = CompareTypes(typ2, typ1);
                        if(!ok)
                         {
                            SimpleWarning("expr is not of type BOOLEAN",
                                             expr->GetPE(2)->minpos,
                                             (int) (expr->GetPE(2)->maxpos - expr->GetPE(2)->minpos +1),
                                             se_err->line);
                            SimpleWarningCont(" --->");
                            typ2->PrintType_nl();
                            
                         }
			delete typ1; typ1 = NULL;
                        delete typ2; typ2 = NULL;
                        se_err = expr->GetSE(5);
                        type = CreateType(expr->GetPE(4));
			typ1 = CreateType(expr->GetPE(6));
			ok = CompareTypes(typ1 ,type);
                        if(!ok)
                         {
                            SimpleWarning("IF...THEN exprs  are of different types",
                                           se_err->pos, se_err->leng, se_err->line);
                            SimpleWarningCont(" IF expr type:\n ");
                            type->PrintType_nl();
                            SimpleWarningCont(" ELSE expr type:\n ");
                            typ1->PrintType_nl();
                           
                         }
                        delete typ1; typ1 = NULL;
                        
                        return type;
	 case CASE    :  type = GetType_CaseRA_List(expr->GetPE(2));  // CASE IBL CaseRA_list IBR
                         return type;
	 case LET     : Handle_expr_list(expr->GetPE(2));  // LET IBL expr_list IN IBL expr IBR
                        type = CreateType(expr->GetPE(4));
                        return type; 
	 case ENABLED : se_err = expr->GetSE(2);   // ENABLED RBR_ON ActionExpr RBR_OFF
                        type = new TypeComp(BOOLEAN_type);
			typ1 = CreateType(expr->GetPE(3));
			ok = CompareTypes(typ1, type);
                        if(!ok)
                         {
                           SimpleWarning("expr is not of type BOOLEAN",
                           expr->GetPE(3)->minpos,
                           (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos +1),
                            se_err->line);
                           SimpleWarningCont(" --->");
                           typ1->PrintType_nl();
                         
                         } 
                        delete typ1; typ1 = NULL;
			return type;
	 case UNCHANGED : type = new TypeComp(BOOLEAN_type);  // UNCHANGED RBR_ON ID_csl RBR_OFF
			  return type;
	 case LT     : se_err = expr->GetSE(1);   // LT expr GT '_' RBR_ON ID_csl RBR_OFF
                       type = new TypeComp(BOOLEAN_type);
		       typ1 = CreateType(expr->GetPE(2));
		       ok = CompareTypes(typ1, type);
                       if(!ok)
                        {
                          SimpleWarning("expr is not of type BOOLEAN",
                          expr->GetPE(2)->minpos,
                          (int) (expr->GetPE(2)->maxpos - expr->GetPE(2)->minpos + 1),
                           se_err->line);
                          SimpleWarningCont(" --->");
                          typ1->PrintType_nl();
                          
                        }
                       delete typ1; typ1 = NULL;
		       return type;
	 case _HENCEFORTH_ :        
		       if(expr->GetMaxIndex() == 3)  // _HENCEFORTH_ RBR_ON expr RBR_OFF
		       { se_err = expr->GetSE(2);
                         type = new TypeComp(BOOLEAN_type);
			 typ1 = CreateType(expr->GetPE(3));
			 ok = CompareTypes(typ1, type);
                         if(!ok)
                          {
                           SimpleWarning("expr is not of type BOOLEAN",
                           expr->GetPE(3)->minpos,
                           (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos + 1),
                            se_err->line);
                           SimpleWarningCont(" --->");
                           typ1->PrintType_nl();
                         
                          }
                         delete typ1; typ1 = NULL;
			 return type;
		       }
		       else    // _HENCEFORTH_ '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
		       { se_err = expr->GetSE(5);
                         type = new TypeComp(BOOLEAN_type);
			 typ1 = CreateType(expr->GetPE(6));
			 ok = CompareTypes(typ1, type);
                         if(!ok)
                          {
                           SimpleWarning("expr is not of type BOOLEAN",
                           expr->GetPE(6)->minpos,
                           (int) (expr->GetPE(6)->maxpos - expr->GetPE(6)->minpos + 1),
                           se_err->line);
                           SimpleWarningCont(" --->");
                           typ1->PrintType_nl();
                        
                          } 
                         delete typ1; typ1 = NULL;
			 return type;
		       }
	 case _EVENTUALLY_ :
		       if(expr->GetMaxIndex() == 3)  // _EVENTUALLY_ RBR_ON expr RBR_OFF
		       { se_err = expr->GetSE(2);
                         type = new TypeComp(BOOLEAN_type);
			 typ1 = CreateType(expr->GetPE(3));
			 ok = CompareTypes(typ1, type);
                         if(!ok)
                          {
                            SimpleWarning("expr is not of type BOOLEAN",
                            expr->GetPE(3)->minpos,
                            (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos +1),
                            se_err->line);
                            SimpleWarningCont(" --->");
                            typ1->PrintType_nl();
                          
                          }
                         delete typ1;  typ1 = NULL;
			 return type;
		       }
		       else     // _EVENTUALLY_  '[' State FcnExpr ']' RBR_ON ActionExpr RBR_OFF
		       { se_err = expr->GetSE(5);
                         type = new TypeComp(BOOLEAN_type);
			 typ1 = CreateType(expr->GetPE(6));
			 ok = CompareTypes(typ1, type);
                         if(!ok)
                          {
                            SimpleWarning("expr is not of type BOOLEAN",
                            expr->GetPE(6)->minpos,
                            (int) (expr->GetPE(6)->maxpos - expr->GetPE(6)->minpos + 1),
                            se_err->line);
                            SimpleWarningCont(" --->");
                            typ1->PrintType_nl();
                           
                          } 
                         delete typ1; typ1 = NULL;
			 return type;
		       }
	  case _BOLD_EXISTS_ :      // _BOLD_EXISTS_ normalID_csl DOUBLECOLON IBL expr IBL
                         se_err = expr->GetSE(3);
			 type = new TypeComp(BOOLEAN_type);
			 typ1 = new TypeComp(JOKER_type);
                         
			 dummy = SetType_normalID_csl(expr->GetPE(2),typ1);
			 typ2 = CreateType(expr->GetPE(4));
			 ok = CompareTypes(typ2, type);
                         if(!ok)
                          {
                            SimpleWarning("expr is not of type BOOLEAN",
                            expr->GetPE(4)->minpos,
                            (int) (expr->GetPE(4)->maxpos - expr->GetPE(4)->minpos +1),
                            se_err->line);
                            SimpleWarningCont(" --->");
                            typ2->PrintType_nl();
                           
                          }
                         delete typ2; typ2 = NULL;
                         
			 return type;
	  case WF  :   // WF '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF

	  case SF  :  se_err = expr->GetSE(5);   // SF '[' StateFcnExpr ']' RBR_ON ActionExpr RBR_OFF
                      type = new TypeComp(BOOLEAN_type);
		      typ1 = CreateType(expr->GetPE(6));
   
		      ok = CompareTypes(typ1, type);
                      if(!ok)
                       {
                        SimpleWarning("expr is not of type BOOLEAN",
                          expr->GetPE(6)->minpos,
                          (int) (expr->GetPE(6)->maxpos - expr->GetPE(6)->minpos + 1),
                          se_err->line);
                        SimpleWarningCont(" --->");
                        typ1->PrintType_nl(); 
                       
                       } 
	              delete typ1; typ1 = NULL;
                      return type;
	  default : assert(0);
	  } // Ende von Terminal-Switch
   case NonTerm :
	   pe = expr->GetPE(1);
	   nt = pe->GetNonTermType();
	   switch(nt)
	   {
	   case texpr : ps_help = expr->GetExpr(2);
	      switch(ps_help->GetSymbolType())
	      {
	      case Terminal :
		if(ps_help->GetSymbElem()->tokenID == TUPELSELECTOR)   // expr TUPELSELECTOR
		    { // Tupelselector ermitteln !
                      se_err = expr->GetSE(2);
		      number = GetSelector(expr->GetSE(2));
		      typ1 = CreateType(pe);
		      type = GetTUPLE(typ1,number);
		      if(type != NULL)
                       { typ2 = type->Copy();
                         delete typ1; typ1 = NULL; 
                         return typ2;
                       }
                      else
		       { SimpleWarning("expr is not of type TUPLE or has inadequate selector",
                                          pe->minpos,
                                          (int) (pe->maxpos - pe->minpos + 1),
                                          se_err->line);
                         SimpleWarningCont(" --->");
                         typ1->PrintType_nl(); 
                         
                         delete typ1; typ1 = NULL;
                         type = new TypeComp(UNKNOWN_type);
			 return type;
		       }
		     }		    
		else if(ps_help->GetSymbElem()->tokenID == DOT)  // expr DOT componentID
		    { // Record-Komponente ermittteln

               
                      se = expr->GetPE(3)->GetSE(1);
                      se_err = expr->GetSE(2);
		      typ1 = CreateType(expr->GetPE(1));

              
                      type = GetRECORD(typ1, se->Getlexem());
                      if(type != NULL)
                       {
                         typ2 = type->Copy();
                         delete typ1; typ1 = NULL;
                         return typ2;
                       }
		       
                      else 
		      { 
                        
                        SimpleWarning("expr is not of type RECORD or has not this component",
                                        expr->GetPE(1)->minpos,
                                        (int) (expr->GetPE(1)->maxpos - expr->GetPE(1)->minpos +1),
                                        se_err->line);
                        SimpleWarningCont(" --->");
                        typ1->PrintType_nl();
                       
                        delete typ1; typ1 = NULL;
                        type = new TypeComp(UNKNOWN_type);
		        return type;
		      }
                     
		    }
		else if(ps_help->GetSymbElem()->tokenID == PRIME)  // expr PRIME
		    { type = CreateType(pe);
		      return type;
		    }
		else if(ps_help->GetSymbElem()->tokenID == _LEADSTO_)  // expr _LEADSTO_ expr
		    { type = new TypeComp(BOOLEAN_type);
		      typ1 = CreateType(pe);
		      typ2 = CreateType(expr->GetPE(3));
                      se_err = expr->GetSE(2);
		      ok = CompareTypes(typ1 , type);
                      if(!ok)
                       {
                         SimpleWarning("expr is not of type BOOLEAN",
                                    pe->minpos,
                                    (int) (pe->maxpos - pe->minpos + 1),
                                    se_err->line);
                         SimpleWarningCont(" --->");
                         typ1->PrintType_nl();
                       
                       } 
		      ok = CompareTypes(typ2 , type);
                      if(!ok)
                       {
                         SimpleWarning("expr is not of type BOOLEAN",
                                expr->GetPE(3)->minpos,
                                (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos + 1),
                                 se_err->line);
                         SimpleWarningCont(" --->");
                         typ2->PrintType_nl();
                       
                       }   
                      delete typ1; typ1 = NULL;
                      delete typ2; typ2 = NULL;
		      return type;
		    }
	      case NonTerm : 
		   if((ps_help->GetParseExpr()->GetNonTermType() == tboolRel) ||
		      (ps_help->GetParseExpr()->GetNonTermType() == tboolRelIB))
                     // expr boolRel expr 
                     // expr boolRelIB IBL expr
		   { se_err = expr->GetPE(2)->GetSE(1);
                     type = new TypeComp(BOOLEAN_type);
		     typ1 = CreateType(pe);
		     typ2 = CreateType(expr->GetPE(3));
		     ok = CompareTypes(typ1 , type);
                     if(!ok)
                      {
                       SimpleWarning("expr is not of type BOOLEAN",
                                   pe->minpos,
                                   (int) (pe->maxpos - pe->minpos + 1),
                                   se_err->line);
                       SimpleWarningCont(" --->");
                       typ1->PrintType_nl();
                       
                      } 
		     ok = CompareTypes(typ2 , type);
                     if(!ok)
                      {
                        SimpleWarning("expr is not of type BOOLEAN",
                                  expr->GetPE(3)->minpos,
                                  (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos +1),
                                  se_err->line);
                        SimpleWarningCont(" --->");
                        typ2->PrintType_nl();
                     
                      } 
                     delete typ1; typ1 = NULL;
                     delete typ2; typ2 = NULL;
		     return type;
		   }
		   else if(ps_help->GetParseExpr()->GetNonTermType() == tBinRel)  // expr BinRel expr
		   { se_err = expr->GetPE(2)->GetSE(1);
                     type = new TypeComp(BOOLEAN_type);
		     typ1 = CreateType(pe);
		     typ2 = CreateType(expr->GetPE(3));
		     ok = CompareTypes(typ2, typ1,1);
                     if(!ok)
                      {
                         SimpleWarning("incompatible types",
                           se_err->pos,se_err->leng,se_err->line);
                         SimpleWarningCont(" type on the right side:\n ");
                         typ2->PrintType_nl();
                         SimpleWarningCont(" type on the left side:\n ");
                         typ1->PrintType_nl();
                       
                      }
                     delete typ1; typ1 = NULL;
                     delete typ2; typ2 = NULL;
		     return type;
		   }
		   else if(ps_help->GetParseExpr()->GetNonTermType() ==
			   tisOrNotElementOf)   // expr isOrNotElementOf SetExpr
		   { type = new TypeComp(BOOLEAN_type);
		     se_err =  expr->GetPE(2)->GetSE(1);

                      
                     typ1 = CreateType(pe);
		     typ2 = CreateType(expr->GetPE(3));
                     typ3 = GetElementofType(typ2);
                     if(se_err->tokenID == _ELEMENT_OF_)
                      { 
                         ok = CompareTypes(typ1,typ3,1);
                         if(!ok)
                          {
                           SimpleWarning("left expr is not Element of right expr",
                                          se_err->pos,se_err->leng,se_err->line);
                           SimpleWarningCont(" type on the right side:\n ");
                           typ2->PrintType_nl();
                           SimpleWarningCont(" type on the left side:\n ");
                           typ1->PrintType_nl();
                         
                           
                          }  
                       };
                     delete typ1; typ1 = NULL;
                     delete typ2; typ2 = NULL;
                     delete typ3; typ3 = NULL;
		     return type;
		    }
		   else if(ps_help->GetParseExpr()->GetNonTermType() ==
			   tBinOp)      // expr BinOp expr
		   { se_err = expr->GetPE(2)->GetSE(1); 
                     type = new TypeComp(NATNUM_type);
		     typ1 = CreateType(pe);
		     typ2 = CreateType(expr->GetPE(3));
                     
		     ok = CompareTypes(typ1 , type);
                     if(!ok)
                       {
                         SimpleWarning("expr not of type NATNUM",
                                       pe->minpos,
                                       (int) (pe->maxpos - pe->minpos + 1),
                                        se_err->line);
                         SimpleWarningCont(" --->");
                         typ1->PrintType_nl();
                         
                       }
		     ok = CompareTypes(typ2 , type);
                     if(!ok)
                      {
                        SimpleWarning("expr not of type NATNUM",
                                     expr->GetPE(3)->minpos,
                                    (int) (expr->GetPE(3)->maxpos - expr->GetPE(3)->minpos + 1),
                                     se_err->line);
                        SimpleWarningCont(" --->");
                        typ2->PrintType_nl();
                        
                      }
                     delete typ1; typ1 = NULL;
                     delete typ2; typ2 = NULL;
		     return type;
		   }
	      default : assert(0);

		  } // Ende von expr-switch
	   case tSetExpr: if(expr->GetMaxIndex() == 2)    // SetExpr setRel SetExpr
			  { se_err = expr->GetPE(2)->GetSE(1);
                            type = new TypeComp(BOOLEAN_type);
			    typ1 = CreateType(pe);
			    typ2 = CreateType(expr->GetPE(3));
                            typ3 = GetElementofType(typ1);
			    ok = CompareTypes(typ2 , typ3,1);  // ist typ2 Element von typ1 ?
                            if(!ok)
                             {
                              SimpleWarning("right expr is not SUBSET type of right expr",
                                        se_err->pos,se_err->leng,se_err->line);
                              SimpleWarningCont(" type on the right side:\n ");
                              typ2->PrintType_nl();
                              SimpleWarningCont(" type on the left side:\n ");
                              typ1->PrintType_nl();
                           
                             } 
                            delete typ1; typ1 = NULL;
                            delete typ2; typ2 = NULL;
                            delete typ3; typ3 = NULL;
			    return type; 
			  }
			  else   // SetExpr
			  { 
                            type = CreateType(pe);
                            
			    return type;
			  }
	   case tStringExpr : type = new TypeComp(STRING_type);  // StringExpr
			      return type;
	   case tprefixLines : type = GetType_prefixLines(pe);  // prefixLines 
                               return type;
	   default :  assert(0);
	      } // Ende von NonTerm Switch
	 
	  
 
   default : assert(0);
	 
  } // Ende von aeserem switch
} // Ende der Routine

