MODULE BankAction 
(* originally from Lamports 9/91 TLA+ text, adapted to            *)
(* 2/92 and to Dortmund-Syntax: Mester, 6.7.92; Geikowski 24.5.93 *)

IMPORT BankInterface;

PARAMETERS  ledger  : Ledger STATE FUNCTION ;
            pending : SET [ATM -> PendingT] STATE FCN ;

INCLUDE Bank WITH Name <- Name;

CONSTANTS
  PendingT  ==  ATMMessage :cup: [[req : {"None"}]]
  ;

PREDICATES
  BInit   == /\ ledger = NewLedger
             /\ pending = [ FCN a :in: ATM :mapsto: [[req :mapsto: "None"]] ]
  ;

  BTypeOK == /\ ledger :in: Ledger
             /\ pending :in: SET [ATM -> PendingT] 
             /\ TypeOK
  ;

ACTIONS
  ReceiveRequest(A : ATM)  == :E: r :in: ATMMessage
                               ::
                                  /\ pending[A].req = "None"
                                  /\ AtoB.Rcv(r,A) 
                                  /\ Unchanged( bToA )
                                  /\ pending' = [CHG pending ; A :mapsto: r]
  ;

  DoRequest(A : ATM)  ==
    /\ pending[A].req # "None"
    /\ pending' = [CHG pending; A :mapsto: [[req :mapsto: "None"]] ]
    /\ Unchanged(aToB)
    /\ LET n  ==  pending[A].nam;
           r  ==  pending[A].req;
           a  ==  pending[A].amt;
       IN  \/ /\ r = "Deposit" 
              /\ ledger' = Deposit(n, a, ledger)
              /\ BtoA.Send("OK", A)
           \/ /\ r = "Withdraw" 
              /\ ledger' = Withdraw(n, a, ledger).2
              /\ BtoA.Send(Withdraw(n, a, ledger).1, A)
           \/ /\ r = "Balance?" 
              /\ ledger' = ledger
              /\ BtoA.Send( Balance(n, ledger), A)
  ;

  NN(A : ATM)  ==  ReceiveRequest(A) \/ DoRequest(A)
  ;

  N  == :E: A :in: ATM :: NN(A)
  ;

THEOREMS
  TypePreserving  ==  /\ Init /\ BInit => BTypeOK
                      /\ BTypeOK /\ N => BTypeOK'
  ;

END
