MODULE EchoActions

  IMPORT Graph,NeedfulThings;

  PARAMETERS Initiator : Node CONSTANT; 
             Marker    ,
             Success   : SET [ Node -> Boolean ];
             ToScan    ,
             QFrom     : SET [ Node -> SUBSET(Node) ];
             Master    : SET [ Node -> Node];
             MBox      : SET [ Node -> QueueType ];

  INCLUDE Queue WITH Type <- Node;

  PREDICATES

    Init == /\ Marker[Initiator] = "true"
            /\ :A: v :in: Node :: v # Initiator => Marker[v] = "false"
            /\ :A: v :in: Node :: /\ Success[v] = "false"
                                  /\ ToScan[v] = Neighbours(v)
                                  /\ QFrom[v] = Neighbours(v)
                                  /\ MBox[v] = Empty
    ;

  ACTIONS

    SendScan(v,w:Node) ==
      /\ Marker[v] = "true"
      /\ w :in: ToScan[v]
      /\ ToScan[v]' = ToScan[v] \ {w}
      /\ MBox[w]' = Insert(MBox[w],v)
      /\ :A: u :in: Node :: /\ u # v => ToScan[u]' = ToScan[u]
                            /\ u # w => MBox[u]' = MBox[u]
      /\ Unchanged(Marker,QFrom,Master,Success)
    ;

    ReceiveFirstScan(v:Node) ==
      /\ Marker[v] = "false"
      /\ ~ IsEmpty(MBox[v])
      /\ Marker[v]' = "true"
      /\ Master[v]' = Head(MBox[v])
      /\ ToScan[v]' = ToScan[v] \ { Head(MBox[v]) }
      /\ QFrom[v]' = QFrom[v] \ { Head(MBox[v]) }
      /\ MBox[v]' = Delete(MBox[v])
      /\ :A: u :in: Node :: v # u => /\ Marker[u]' = Marker[u]
                                     /\ Master[u]' = Master[u]
                                     /\ ToScan[u]' = ToScan[u]
                                     /\ QFrom[u]' = QFrom[u]
                                     /\ MBox[u]' = MBox[u]
      /\ Unchanged(Success)
    ;

    ReceiveSucc(v:Node) ==
      /\ Marker[v] = "true"
      /\ ~ IsEmpty(MBox[v])
      /\ QFrom[v]' = QFrom[v] \ { Head(MBox[v]) }
      /\ MBox[v]' = Delete(MBox[v])
      /\ :A: u :in: Node :: v # u => /\ QFrom[u]' = QFrom[u]
                                     /\ MBox[u]' = MBox[u]
      /\ Unchanged(Marker,ToScan,Master,Success)
    ;

    SendReply(v:Node) ==
      /\ Marker[v] = "true"
      /\ v # Initiator
      /\ QFrom[v] = {}
      /\ Success[v] = "false"
      /\ Success[v]' = "true"
      /\ MBox[ Master[v] ]' = Insert(MBox[ Master[v] ],v)
      /\ :A: u :in: Node :: /\ u # v => Success[u]' = Success[u]
                            /\ u # Master[v] => MBox[u]' = MBox[u]
      /\ Unchanged(Marker,QFrom,Master,ToScan)
    ;

    InitiatorDetectsTermination(v:Node) ==
      /\ Marker[v] = "true"
      /\ v = Initiator
      /\ QFrom[v] = {}
      /\ Success[v] = "false"
      /\ Success[v]' = "true"
      /\ :A: u :in: Node :: u # v => Success[u]' = Success[u]
      /\ Unchanged(Marker,QFrom,Master,ToScan,MBox)
    ;

END
