TEMPORAL MODULE Wire
(* originally from Lamports 9/91 TLA+ text, adapted to             *)
(* 2/92 and to Dortmund-Syntax: Mester, 1.7.92; Geikowski, 18.5.93 *)

EXPORT OPAQUE Type, Init, Send, Rcv;

PARAMETERS Item : Any CONST;    (* The "type" of items that can be sent *)

CONSTANTS
  Boolean == { "true", "false" };
  Type    == [[itm : Item; rdy : Boolean]];
  
PREDICATES
  Init(wire : Type) == /\ wire :in: Type
                       /\ wire.rdy = "true"
  ;
       
ACTIONS
  Send(wire : Type; i : Item) ==
    /\ wire.rdy = "true"
    /\ wire' = [[itm :mapsto: i, rdy :mapsto: "false"]]
  ;
  Rcv(wire : Type; i : Item) ==
    /\ wire.rdy = "false"
    /\ wire.itm = i
    /\ wire' = [[itm :mapsto: i, rdy :mapsto: "true"]]
  ; 

THEOREMS
  TypePreserving(wire : Type)  ==  
    /\ Init(wire) => wire :in: Type
    /\ :A: i :in: Item
        ::
           /\  wire :in: Type /\ Send(wire,i)  =>  wire' :in: Type 
           /\  wire :in: Type /\ Rcv(wire,i)   =>  wire' :in: Type 
  ;
  WireAxiom(wire : Type)  == 
    /\ Init(wire) => :A: i :in: Item :: Enabled( Send(wire,i) )
    /\ wire :in: Type
         => :A: i, j :in: Item
             ::
                /\ ~ ( Enabled( Send(wire,i) ) /\ Enabled( Rcv(wire,j)) )
                /\ Send(wire,i) => (Enabled( Rcv(wire,j) ) )'
                /\ Rcv(wire,i)  => (Enabled( Send(wire,j) ) )'
  ;

END
